# -*- coding: utf-8 -*-

"""This module contains the definition of class `OrganizationalModel`.

This module also contains the implementation of the default mining 
method by Song and van der Aalst [1]_, as a naive approach for
organizational model discovery from event logs.

References
----------
.. [1] Song, M., & van der Aalst, W. M. P. (2008). Towards comprehensive 
   support for organizational mining. *Decision Support Systems*, 46(1), 
   300-317. `<https://doi.org/10.1016/j.dss.2008.07.002>`_
"""
class OrganizationalModel:
    """This class defines an organizational model.

    An organizational model includes the following data structures as
    fundamentals:

        - Resource Group ID (`_rg_id`), resource group ids,
        - Resource Group (`_rg`), a mapping from resource group ids to
          group descriptions,
        - Membership (`_mem`), a mapping from resource group ids to
          member resource ids,
        - Capability (`_cap`), a mapping from resource group ids to
          capable execution modes.

    Attributes
    ----------
    group_number : int
        The number of resource groups in the model.
    resources : frozenset
        All the resources involved in the model.

    Methods
    -------
    add_group(og, exec_modes)
        Add a resource group into the organizational model.
    find_group_ids(r)
        Query the id of groups which contain a resource given its 
        identifier.
    find_groups(r)
        Query the groups which contain a resource given its identifier.
    find_execution_modes(r)
        Query the allowed execution modes of a resource given its 
        identifier.
    find_candidate_groups(exec_mode)
        Query the capable groups (i.e., groups that can perform the 
        execution mode according to the model) given an execution mode.
    find_all_groups()
        Return all resource groups involved in the model.
    find_group_members(rg_id)
        Query the group members given a group identified by its id.
    find_group_execution_modes(rg_id)
        Query the capable execution modes given a group identified by 
        its id.
    find_all_execution_modes()
        Return all execution modes involved in the model.
    to_file_csv(f)
        Export the organizational model to an external CSV file (with no 
        header line).
    from_file_csv(f)
        Import an organizational model from an external CSV file.

    Notes
    -----
    Methods related to the query on an organizational model are also 
    defined.
    """

    def __init__(self):
        self._rg_id = -1
        self._rg = dict()
        self._mem = dict()
        self._cap = dict()

        from collections import defaultdict
        self._rmem = defaultdict(set)
        self._rcap = defaultdict(set)
        

    def add_group(self, og, exec_modes):
        """Add a resource group into the organizational model.

        Parameters
        ----------
        og : iterator
            Ids of resources to be added as a resource group.
        exec_modes : iterator, list or dict of lists
            Execution modes corresponding to the group to be added.

        Returns
        -------

        Raises
        ------
        TypeError
            If the parameter type for `exec_modes` is unexpected.
        """
        if type(exec_modes) is list:
            # no refinement applied
            self._rg_id += 1
            self._rg[self._rg_id] = ''

            self._mem[self._rg_id] = set()
            for r in og:
                self._mem[self._rg_id].add(r)
                self._rmem[r].add(self._rg_id)

            self._cap[self._rg_id] = list()
            for m in exec_modes:
                self._cap[self._rg_id].append(m)
                self._rcap[m].add(self._rg_id)
        elif type(exec_modes) is dict:
            # refinement applied
            # TODO: consider deprecating
            for subog, subm in exec_modes.items():
                if subog not in self._mem.values():
                    self._rg_id += 1
                    self._rg[self._rg_id] = ''

                    self._mem[self._rg_id] = set()
                    for r in subog:
                        self._mem[self._rg_id].add(r)
                        self._rmem[r].add(self._rg_id)

                    self._cap[self._rg_id] = list()
                    for m in subm:
                        self._cap[self._rg_id].append(m)
                        self._rcap[m].add(self._rg_id)
        else:
            raise TypeError('Invalid type for parameter `{}`: {}'.format(
                'exec_modes', type(exec_modes)))
            

    @property
    def group_number(self):
        """Return the number of resource groups of the model.

        Parameters
        ----------

        Returns
        -------
        int
            Number of resource groups.
        """
        return len(self._rg)


    @property
    def resources(self):
        """Return all the resources involved in the model.

        Parameters
        ----------

        Returns
        -------
        frozenset
            Set of all resources.
        """
        return frozenset(self._rmem.keys())


    def find_group_ids(self, r):
        """Query the id of groups which contain a resource given its 
        identifier.

        Parameters
        ----------
        r : str
            The identifier of a given resource.

        Returns
        -------
        list of ints
            Id of the groups to which the queried resource belongs.
        """
        return list(self._rmem[r])


    def find_groups(self, r):
        """Query the groups which contain a resource given its 
        identifier.

        Parameters
        ----------
        r : str
            The identifier of a given resource.

        Returns
        -------
        list of frozensets
            Groups to which the queried resource belongs.
        """
        return [frozenset(self._mem[rg_id]) 
            for rg_id in self._rmem[r]]


    def find_execution_modes(self, r):
        """Query the allowed execution modes of a resource given its 
        identifier.

        Parameters
        ----------
        r : str
            The identifier of a given resource.

        Returns
        -------
        list of 3-tuples
            Allowed execution modes specific to the queried resource.
        """
        resource_cap = set()
        for rg_id in self._rmem[r]:
            resource_cap.update(set(self._cap[rg_id]))
        return list(resource_cap)
    

    def find_candidate_groups(self, exec_mode):
        """Query the capable groups (i.e., groups that can perform the 
        execution mode according to the model) given an execution mode.

        Parameters
        ----------
        exec_mode : 3-tuple
            The execution mode given.

        Returns
        -------
        list of frozensets
            Groups of resources which are capable of this execution mode.
        """
        return [frozenset(self._mem[rg_id]) 
            for rg_id in self._rcap[exec_mode]]


    def find_all_groups(self):
        """Return all resource groups involved in the model.

        Parameters
        ----------

        Returns
        -------
        list of 2-tuples
            Ids and member resources of the groups.
        """
        return [(rg_id, frozenset(self._mem[rg_id])) 
            for rg_id in self._rg.keys()]


    def find_group_members(self, rg_id):
        """Query the capable execution modes given a group identified by 
        its id.

        Parameters
        ----------
        rg_id : int
            The given group id.

        Returns
        -------
        frozenset
            Set of member resources.
        """
        return frozenset(self._mem[rg_id])


    def find_group_execution_modes(self, rg_id):
        """Query the capable execution modes given a group identified by 
        its id.

        Parameters
        ----------
        rg_id : int
            The given group id.

        Returns
        -------
        list of 3-tuples
            Allowed execution modes specific to the given group.
        """
        return self._cap[rg_id]
    

    def find_all_execution_modes(self):
        """Return all execution modes involved in the model.

        Parameters
        ----------

        Returns
        -------
        list of 3-tuples
            Execution modes.
        """
        all_modes = list()
        for modes in self._cap.values():
            all_modes.extend(modes)
        return all_modes
    

    def to_file_csv(self, f):
        """Export the organizational model to an external CSV file (with
        no header line).

        Parameters
        ----------
        f : File object
            A destination CSV file to be written.

        Returns
        -------

        Notes
        -----
        Data exchange format in the CSV file (by each line):

        * No header line; 
        * 3 columns: Group ID, Member resource IDs, Execution modes
        * Values of member resource IDs and execution modes are separated
          by semicolons ``;`` within the columns;
        * A vertical bar ``|`` is used as the separator for values of the
          three types within an execution mode.
        
        An example is given as following:

        ======== =========================== ========================
        GroupID1 Resource ID 1;Resource ID 2 CT1|AT2|TT1;CT.0|AT2|TT3
        -------- --------------------------- ------------------------
        GroupID2 Resource ID 3;Resource ID 4 CT1|AT3|TT1;CT.0|AT2|TT4
        ======== =========================== ========================

        See Also
        --------
        OrganizationalModel.from_file_csv
        """
        from csv import writer
        writer = writer(f)

        rows = list()
        for rg_id in sorted(self._rg.keys()):
            str_rg_id = str(rg_id)
            str_members = ';'.join(sorted(str(r) for r in self._mem[rg_id]))
            if len(self._cap) > 0:
                str_exec_modes = ';'.join(sorted('|'.join(str(t) for t in mode)
                    for mode in self._cap[rg_id]))
            else:
                str_exec_modes = ''

            rows.append([str_rg_id, str_members, str_exec_modes])

        writer.writerows(rows)


    @classmethod
    def from_file_csv(cls, f):
        """Import an organizational model from an external CSV file.

        Parameters
        ----------
        f: File object
            A sourced CSV file to be read.

        Returns
        -------
        OrganizationalModel
            The imported organizational model.

        Notes
        -----
        Data exchange format in the CSV file (by each line):

        * No header line; 
        * 3 columns: Group ID, Member resource IDs, Execution modes
        * Values of member resource IDs and execution modes are separated
          by semicolons ``;`` within the columns;
        * A vertical bar ``|`` is used as the separator for values of the
          three types within an execution mode.
        
        An example is given as follows:

        ======== =========================== ========================
        GroupID1 Resource ID 1;Resource ID 2 CT1|AT2|TT1;CT.0|AT2|TT3
        -------- --------------------------- ------------------------
        GroupID2 Resource ID 3;Resource ID 4 CT1|AT3|TT1;CT.0|AT2|TT4
        ======== =========================== ========================

        See Also
        --------
        OrganizationalModel.to_file_csv
        """
        from csv import reader
        om_obj = cls()
        for row in reader(f):
            if not row:
                continue

            group = row[1].split(';')
            exec_modes = list()
            for str_mode in row[2].split(';'):
                exec_modes.append(tuple(str_mode.split('|')))
            om_obj.add_group(group, exec_modes)
        return om_obj


def default_mining(rl):
    """The default mining method.

    Parameters
    ----------
    rl : DataFrame
        A resource log.

    Returns
    -------
    ogs : list of frozensets
        A list of resource groups.

    Notes
    -----
    This method does not require using resource profiles.
    """
    print('Applying Default Mining:')
    ogs = list()
    for activity_type, events in rl.groupby('activity_type'):
        ogs.append(frozenset(events['resource']))
    return ogs




# -*- coding: utf-8 -*-

"""This module contains the implementation of graph/network -based 
organizational mining methods, based on the use of graph partitioning 
techniques.
"""
from warnings import warn

def _mja(profiles, n_groups, metric='euclidean'):
    """Apply Metrics based on Joint Activities to discover organizational 
    groups [1]_.

    Parameters
    ----------
    profiles : DataFrame Constructed resource profiles. n_groups : int
        Expected number of resource groups. 
    metric : str, optional, default 'euclidean' 
        Choice of metrics for measuring the distance while calculating
        distance. Defaults to ``'euclidean'``, meaning that euclidean
        distance is used for measuring distance.

    Returns
    -------
    ogs : list of frozensets Discovered resource groups.

    Raises
    ------
    RuntimeError If the specified number of groups could not be
        discovered.

    See Also
    --------
    orgminer.SocialNetworkMiner.joint_activities

    References
    ----------
    .. [1] Van der Aalst, W. M. P., Reijers, H. A., & Song, M. (2005).
       Discovering social networks from event logs. *Computer Supported
       Cooperative Work (CSCW)*, 14(6), 549-593.
       `<https://doi.org/10.1007/s10606-005-9005-9>`_
    """
    print('Applying graph/network-based MJA:')
    from orgminer.SocialNetworkMiner.joint_activities import distance
    sn = distance(profiles, metric=metric, convert=True)

    from operator import itemgetter
    edges_sorted = sorted(sn.edges.data('weight'), key=itemgetter(2))
    from networkx import restricted_view
    from networkx import connected_components, number_connected_components
    # search the cut edge using bisection (i.e. binary search)
    lo = 0
    hi = len(edges_sorted)
    while lo < hi:
        mid = (lo + hi) // 2
        sub_sn = restricted_view(
            sn, nodes=[], edges=[(u, v) for u, v, w in edges_sorted[:mid]])
        if number_connected_components(sub_sn) < n_groups:
            lo = mid + 1
        else:
            hi = mid
    sub_sn = restricted_view(
        sn, nodes=[], edges=[(u, v) for u, v, w in edges_sorted[:lo]])
    ogs = list()
    if number_connected_components(sub_sn) == n_groups:
        for comp in connected_components(sub_sn):
            ogs.append(frozenset(comp))
    else:
        raise RuntimeError('Unable to discover specified number of groups.')
    return ogs


def mja(profiles, n_groups, metric='euclidean', search_only=False):
    """Apply Metrics based on Joint Activities to discover organizational 
    groups [1]_.

    Parameters
    ----------
    profiles : DataFrame
        Constructed resource profiles.
    n_groups : int, or list of ints
        Expected number of resource groups, or a list of candidate
        numbers to be determined.
    metric : str, optional, default 'euclidean'
        Choice of metrics for measuring the distance while calculating 
        distance. Defaults to ``'euclidean'``, meaning that euclidean
        distance is used for measuring distance.
    search_only : bool, optional, default False
        A boolean flag indicating whether to search for the number of
        groups only or to perform group discovery based on the search
        result. Defaults to ``False``, i.e., to perform group discovery
        after searching.

    Returns
    -------
    best_k : int
        The suggested selection of number of groups (if `search_only` is
        True).
    list of frozensets
        Discovered resource groups (if `search_only` is False).

    Raises
    ------
    TypeError
        If the parameter type for `n_groups` is unexpected.

    See Also
    --------
    orgminer.SocialNetworkMiner.joint_activities

    References
    ----------
    .. [1] Van der Aalst, W. M. P., Reijers, H. A., & Song, M. (2005).
       Discovering social networks from event logs. *Computer Supported
       Cooperative Work (CSCW)*, 14(6), 549-593.
       `<https://doi.org/10.1007/s10606-005-9005-9>`_
    """
    if type(n_groups) is int:
        return _mja(profiles, n_groups, metric)
    elif type(n_groups) is list and len(n_groups) == 1:
        return _mja(profiles, n_groups[0], metric)
    elif type(n_groups) is list and len(n_groups) > 1:
        from orgminer.OrganizationalModelMiner.utilities import \
            cross_validation_score
        best_k = -1
        best_score = float('-inf')
        for k in n_groups:
            score = cross_validation_score(
                X=profiles, miner=_mja,
                miner_params={
                    'n_groups': k,
                    'metric': metric
                },
                proximity_metric=metric
            )
            if score > best_score:
                best_score = score
                best_k = k

        print('-' * 80)
        print('Selected "K" = {}'.format(best_k))
        if search_only:
            return best_k 
        else:
            return _mja(profiles, best_k, metric)
    else:
        raise TypeError('Invalid type for parameter `{}`: {}'.format(
            'n_groups', type(n_groups)))


def _mjc(el, n_groups, method='threshold'):
    """Apply Metrics based on Joint Cases to discover organizational 
    groups [1]_.

    Parameters
    ----------
    el : DataFrame
        An event log.
    n_groups : int
        Expected number of resource groups.
    method : {'threshold', 'centrality'}, optional, default 'threshold'
        Options for the method to be used for finding graph components. 
        Could be one of the following:

            - 'threshold': using edge thresholding on edges to remove
              links.
            - 'centrality': disconnect nodes with high betweenness 
              centrality, i.e., shortest-path centrality.

    Returns
    -------
    ogs : list of frozensets
        Discovered resource groups.
    
    Raises
    ------
    ValueError
        If the specified method for finding graph components is invalid.

    See Also
    --------
    orgminer.SocialNetworkMiner.joint_cases

    References
    ----------
    .. [1] Van der Aalst, W. M. P., Reijers, H. A., & Song, M. (2005).
       Discovering social networks from event logs. *Computer Supported
       Cooperative Work (CSCW)*, 14(6), 549-593.
       `<https://doi.org/10.1007/s10606-005-9005-9>`_
    """
    print('Applying graph/network-based MJC:')
    from orgminer.SocialNetworkMiner.joint_cases import working_together
    # directed graph
    sn = working_together(el, normalize='resource')

    from networkx import restricted_view
    from networkx import strongly_connected_components as cc
    from networkx import number_strongly_connected_components as num_cc

    if method == 'threshold':
        # Eliminate less-important edges and maintain the stronger ones
        # Casting from DiGraph to Graph needs to be configured manually
        # otherwise NetworkX would approach this in an arbitrary fashion
        from itertools import combinations
        undirected_edge_list = list()
        for pair in combinations(sn.nodes, r=2):
            if (sn.has_edge(pair[0], pair[1]) and 
                sn.has_edge(pair[1], pair[0])):
                undirected_edge_wt = 0.5 * (
                    sn[pair[0]][pair[1]]['weight'] +
                    sn[pair[1]][pair[0]]['weight'])
                undirected_edge_list.append(
                    (pair[0], pair[1], {'weight': undirected_edge_wt}))
            else:
                pass
        sn.clear()
        del sn
        from networkx import Graph
        sn = Graph()
        sn.add_edges_from(undirected_edge_list)
        del undirected_edge_list[:]
        warn('DiGraph casted to Graph.', RuntimeWarning)

        from operator import itemgetter
        edges_sorted = sorted(sn.edges.data('weight'), key=itemgetter(2))
        # search the cut edge using bisection (i.e. binary search)
        lo = 0
        hi = len(edges_sorted)
        while lo < hi:
            mid = (lo + hi) // 2
            sub_sn = restricted_view(
                sn, nodes=[], edges=[(u, v) for u, v, w in edges_sorted[:mid]])
            if num_cc(sub_sn) < n_groups:
                lo = mid + 1
            else:
                hi = mid
        sub_sn = restricted_view(
            sn, nodes=[], edges=[(u, v) for u, v, w in edges_sorted[:lo]])
    elif method == 'centrality':
        # Disconnect particular nodes
        from networkx import betweenness_centrality
        # betweenness centrality can be calculated on directed graphs
        node_centrality = betweenness_centrality(sn, weight='weight')
        from operator import itemgetter
        # sorted the nodes by centrality in a descending order
        nodes_sorted = list(map(itemgetter(0), sorted(
            list(node_centrality.items()), key=itemgetter(1), reverse=True)))
        # search the disconnected node using bisection (i.e. binary search)
        lo = 0
        hi = len(nodes_sorted)
        while lo < hi:
            mid = (lo + hi) // 2
            # discard the connecting links but not the nodes themselves
            nodes_to_disconnect = nodes_sorted[:mid]
            edges_to_disconnect = (
                list(sn.in_edges(nbunch=nodes_to_disconnect)) +
                list(sn.out_edges(nbunch=nodes_to_disconnect)))
            sub_sn = restricted_view(
                sn, nodes=[],
                edges=edges_to_disconnect)
            if num_cc(sub_sn) < n_groups:
                lo = mid + 1
            else:
                hi = mid
        nodes_to_disconnect = nodes_sorted[:lo]
        edges_to_disconnect = (
            list(sn.in_edges(nbunch=nodes_to_disconnect)) +
            list(sn.out_edges(nbunch=nodes_to_disconnect)))
        sub_sn = restricted_view(
            sn, nodes=[], edges=edges_to_disconnect)

    else:
        raise ValueError('Invalid value for parameter `{}`: {}'.format(
            'method', method))

    ogs = list()
    if num_cc(sub_sn) == n_groups:
        for comp in cc(sub_sn):
            ogs.append(frozenset(comp))
    else:
        pass
    return ogs


def mjc(el, n_groups, search_only=False):
    """Apply Metrics based on Joint Cases to discover organizational 
    groups [1]_.

    Parameters
    ----------
    el : DataFrame
        An event log.
    n_groups : int, or list of ints
        Expected number of resource groups, or a list of candidate
        numbers to be determined.
    search_only : bool, optional, default False
        A boolean flag indicating whether to search for the number of
        groups only or to perform group discovery based on the search
        result. Defaults to ``False``, i.e., to perform group discovery
        after searching.

    Returns
    -------
    best_k : int
        The suggested selection of number of groups (if `search_only` is
        True).
    list of frozensets
        Discovered resource groups (if `search_only` is False).

    Raises
    ------
    TypeError
        If the parameter type for `n_groups` is unexpected.

    See Also
    --------
    orgminer.SocialNetworkMiner.joint_cases

    References
    ----------
    .. [1] Van der Aalst, W. M. P., Reijers, H. A., & Song, M. (2005).
       Discovering social networks from event logs. *Computer Supported
       Cooperative Work (CSCW)*, 14(6), 549-593.
       `<https://doi.org/10.1007/s10606-005-9005-9>`_
    """
    if type(n_groups) is int:
        return _mjc(el, n_groups)
    elif type(n_groups) is list and len(n_groups) == 1:
        return _mjc(el, n_groups[0])
    elif type(n_groups) is list and len(n_groups) > 1:
        # TODO: How to evaluate a result from applying MJC?
        raise NotImplementedError
    else:
        raise TypeError('Invalid type for parameter `{}`: {}'.format(
            'n_groups', type(n_groups)))




# -*- coding: utf-8 -*-

"""This module contains the implementation of overlapping graph/network 
-based organizational mining methods, based on the use of community 
detection techniques [1]_.

References
----------
.. [1] Xie, J., Kelley, S., & Szymanski, B. K. (2013). Overlapping
   community detection in networks. *ACM Computing Surveys*, 45(4), 1–35.
   `<https://doi.org/10.1145/2501654.2501657>`_
"""
from deprecated import deprecated
from warnings import warn

def _relabel_nodes_integers(g):
    """A helper function that relabels nodes in a graph/network using 
    consecutive integer ids starting from 0. The relabeled ids are 
    ordered following the original alphabetical order as the original 
    node labels.

    Parameters
    ----------
    g : NetworkX Graph or DiGraph
        The original network of which the nodes are to be relabeled.

    Returns
    -------
    NetworkX Graph or DiGraph
        A new network with nodes relabeled.
    rev_mapping : dict
        A mapping from the new node ids (integers) to the original node 
        labels.

    See Also
    --------
    networkx.relabel.relabel_nodes
    """
    mapping = dict()
    rev_mapping = dict()
    for i, node in enumerate(sorted(g.nodes())):
        mapping[node] = i
        rev_mapping[i] = node
    from networkx import relabel_nodes
    return relabel_nodes(g, mapping), rev_mapping


def _extended_modularity(g, cover):
    """A helper function that calculates the extended modularity [1]_ 
    given a graph/network and a cover (i.e., a set of communities).

    Parameters
    ----------
    g : NetworkX Graph
        A graph/network.
    cover : list of sets
        A cover (a set of communities of nodes).
    
    Returns
    -------
    float
        The result extended modularity value.

    Notes
    -----
    The data attribute name of edge weight defaults to ``'weight'``, as
    in NetworkX.

    References
    ----------
    .. [1] Shen, H., Cheng, X., Cai, K., & Hu, M. B. (2009). Detect
       overlapping and hierarchical community structure in networks.
       *Physica A: Statistical Mechanics and its Applications*, 388(8),
       1706-1712. `<https://doi.org/10.1016/j.physa.2008.12.021>`_
    """
    from collections import defaultdict
    node_membership = defaultdict(set)
    # identify membership
    for i, community in enumerate(cover):
        for node in community:
            node_membership[node].add(i)

    # calculate extended modularity by iterating over every pair of
    # distint nodes within a community
    eq = 0.0
    m = sum([wt for (v, w, wt) in g.edges.data('weight')])
    from itertools import combinations
    for community in cover:
        for v, w in combinations(community, 2):
            Ov = len(node_membership[v])
            Ow = len(node_membership[w])
            Avw = g.edges[v, w]['weight'] if g.has_edge(v, w) else 0
            kv = g.degree(v, 'weight')
            kw = g.degree(w, 'weight')
            eq += (1.0 / (Ov * Ow) * (Avw - kv * kw / (2 * m)))
    return eq


def clique_percolation(profiles, metric='euclidean'):
    """Apply a clique percolation technique [1]_ for detecting 
    communities and thus to discover organizational groups.

    Parameters
    ----------
    profiles : DataFrame
        Constructed resource profiles.
    metric : str, optional, default 'euclidean'
        Choice of metrics for measuring the distance while calculating 
        distance. Defaults to ``'euclidean'``, meaning that euclidean
        distance is used for measuring distance.

    Returns
    -------
    ogs : list of frozensets
        Discovered resource groups.
    
    Notes
    -----
    Using this function relies on an external tool, CFinder [1]_. This 
    function provides merely the interface between OrgMiner and the 
    external tool.
    
    See Also
    --------
    orgminer.SocialNetworkMiner.joint_activities
    scipy.spatial.distance

    References
    ----------
    .. [1] Palla, G., Derenyi, I., Farkas, I., & Vicsek, T. (2005).
       Uncovering the overlapping community structure of complex networks
       in nature and society. *Nature*, 435(7043), 814–818.
       `<https://doi.org/10.1038/nature03607>`_
    """
    print('Applying graph/network-based Clique Percolation Method ' + 
        '(overlapping community detection using CFinder):')
    # build network from profiles
    from orgminer.SocialNetworkMiner.joint_activities import distance
    sn = distance(profiles, metric=metric, convert=True)

    # step 0. Relabel nodes
    sn, rev_node_relabel_mapping = _relabel_nodes_integers(sn)

    # step 1. Distinguish the isolated nodes
    from networkx import isolates
    original_isolates = list(isolates(sn))
    if len(original_isolates) > 0:
        warn('There exist {} isolated nodes in the network.'.format(
            len(original_isolates)), RuntimeWarning)

    # step 2. Export network as edgelist
    from networkx import write_weighted_edgelist
    with open('tmp_sn.edgelist', 'wb') as f:
        write_weighted_edgelist(sn, f)
    print('Network exported to "tmp_sn.edgelist".')
    print('Use the external tool CFinder to discover communities:')

    # step 3. Run community detection applying CFinder
    print('Path to the output directory as input: ', end='')
    dirn_output = input()

    # step 4. Derive organizational groups from the detection results
    # Note: since CFinder produces a set of results varied by num. of cliques
    from collections import defaultdict
    best_fn = None
    best_eq = float('-inf')
    fn_cnt = 0
    solution = None
    from os import listdir, path
    for n in listdir(dirn_output):
        if path.isdir(path.join(dirn_output, n)) and n.startswith('k='):
            fn_cnt += 1
            fn_communities = path.join(dirn_output, n, 'communities')
            cnt = -1
            groups = defaultdict(set)
            with open(fn_communities, 'r') as f:
                for line in f:
                    if not (line == '' or line.startswith('#')):
                        cnt += 1
                        for label in line.split(':')[-1].strip().split():
                            groups[cnt].add(int(label))
            for i, iso_node in enumerate(original_isolates):
                groups['ISOLATE #{}'.format(i)].add(iso_node)
            eq = _extended_modularity(sn, list(groups.values()))
            if eq > best_eq:
                best_fn = fn_communities
                best_eq = eq
                solution = groups
    
    print('Detected communities imported from directory "{}":'.format(
        dirn_output))
    print('Best solution "{}" selected from {} candidates:'.format(
        best_fn, fn_cnt))
    print('{} organizational groups discovered.'.format(len(solution)))
    # restore labels
    ogs = list()
    for cover in solution.values():
        ogs.append(frozenset({rev_node_relabel_mapping[x] for x in cover}))
    return ogs


@deprecated(reason='This method requires a dependency unabled to be resolved.')
def link_partitioning(profiles, n_groups, metric='euclidean'):
    """Apply a link partitioning technique [1]_ for detecting communities 
    and thus to discover organizational groups [2]_.

    Parameters
    ----------
    profiles : DataFrame
        Constructed resource profiles.
    n_groups : int
        Expected number of resource groups.
    metric : str, optional, default 'euclidean'
        Choice of metrics for measuring the distance while calculating 
        distance. Defaults to ``'euclidean'``, meaning that euclidean
        distance is used for measuring distance.

    Returns
    -------
    list of frozensets
        Discovered resource groups.

    Notes
    -----
    This function is an implementation of the method proposed in [2]_.

    See Also
    --------
    orgminer.SocialNetworkMiner.joint_activities
    scipy.spatial.distance

    References
    ----------
    .. [1] Evans, T. S., & Lambiotte, R. (2010). Line graphs of weighted
       networks for overlapping communities. *The European Physical
       Journal B*, 77(2), 265–272.
       `<https://doi.org/10.1140/epjb/e2010-00261-8>`_
    .. [2] Appice, A. (2017). Towards mining the organizational
       structure of a dynamic event scenario. *Journal of Intelligent
       Information Systems*, 1–29.
       `<https://doi.org/10.1007/s10844-017-0451-x>`_
    """
    print('Applying graph/network-based link partitioning method ' + 
        '(overlapping community detection):')
    raise NotImplementedError # TODO: dependency on louvain unresolved
    # build network from profiles
    from orgminer.SocialNetworkMiner.joint_activities import distance
    sn = distance(profiles, metric=metric, convert=True)
    from orgminer.SocialNetworkMiner.utilities import select_edges_by_weight
    sn = select_edges_by_weight(sn, low=0) # Appice's setting
    sn = select_edges_by_weight(sn, percentage='+0.75') # Appice's setting

    # step 1. Build the linear network using the original network
    # distinguish the isolated nodes in the original network first
    # store as a Pajek .net format file as intermediate
    from networkx import isolates
    original_isolates = list(isolates(sn))
    if len(original_isolates) > 0:
        warn('There exist {} isolated nodes in the network.'.format(
            len(original_isolates)), RuntimeWarning)
        print('When using an external tool to discover '
            'communities from the linear network, you should specify the'
            ' target number of communities as:\n\tN\' = N - {},'.format(
                len(original_isolates)), end=' ')
        print('where N is the actual target number to be obtained.')

    tmp_file_path = '.linear_graph.tmp'
    edges = sorted(list(sn.edges.data('weight')))
    with open(tmp_file_path, 'w') as f_pajek_net:
        # write header
        f_pajek_net.write('*Vertices {}\n'.format(len(edges)))

        # write nodes in LN
        for i, e in enumerate(edges):
            # "><" is used as deliminator to distinguish the original nodes
            f_pajek_net.write('{} "{}><{}"\n'.format(i + 1, 
                str(e[0]), str(e[1])))

        # write header
        f_pajek_net.write('*arcs\n')

        # write edges in LN
        from itertools import combinations
        for i, j in combinations(range(len(edges)), 2):
            ei = edges[i]
            ej = edges[j]
            joint = None
            if ei[0] == ej[0] or ei[0] == ej[1]:
                joint = ei[0]
            elif ei[1] == ej[0] or ei[1] == ej[1]:
                joint = ei[1]

            if joint is not None:
                # i -> j
                w_ij = (ei[2] 
                    / (sn.degree(nbunch=joint, weight='weight') - ej[2]))
                # i <- j
                w_ji = (ej[2] 
                    / (sn.degree(nbunch=joint, weight='weight') - ei[2]))

                # precision set to 1e-9
                f_pajek_net.write('{} {} {:.9f}\n'.format(i + 1, j + 1, w_ij))
                f_pajek_net.write('{} {} {:.9f}\n'.format(j + 1, i + 1, w_ji))

    # step 2. Run Louvain algorithm to discover communities
    # convert the graph to igraph format
    from igraph import Graph as iGraph
    ln_igraph = iGraph.Read_Pajek(tmp_file_path)
    from os import remove
    remove(tmp_file_path)

    import louvain
    louvain.set_rng_seed(0)
    optimiser = louvain.Optimiser()
    # search the resolution parameter using bisection
    lo = 0.5
    hi = 1.0
    eps = 0.01
    while (lo + hi) / 2 > eps:
        mid = (lo + hi) / 2
        partition = louvain.RBConfigurationVertexPartition(
            ln_igraph, weights='weight', resolution_parameter=mid)
        diff_inc = 1
        while diff_inc > 0:
            diff_inc = optimiser.optimise_partition(partition)
        ln_communities = partition

        if len(ln_communities) == n_groups:
            break
        elif len(ln_communities) < n_groups:
            lo = mid
        else:
            hi = mid

    # step 3. Map communities onto the original network to get the results
    # derive the orgnizational groups
    from collections import defaultdict
    groups = defaultdict(set)
    for i, comm in enumerate(ln_communities):
        for n_idx in comm:
            ln_v = ln_igraph.vs[n_idx]
            groups[i].add(ln_v['id'].split('><')[0])
            groups[i].add(ln_v['id'].split('><')[1])
    for i, iso_node in enumerate(original_isolates):
        groups['ISOLATE #{}'.format(i)].add(iso_node)
    return [frozenset(g) for g in groups.values()]


def local_expansion(profiles, metric='euclidean'):
    """Apply a local expansion technique [1]_ for detecting communities 
    and thus to discover organizational groups.

    Parameters
    ----------
    profiles : DataFrame
        Constructed resource profiles.
    metric : str, optional, default 'euclidean'
        Choice of metrics for measuring the distance while calculating 
        distance. Defaults to ``'euclidean'``, meaning that euclidean
        distance is used for measuring distance.

    Returns
    -------
    list of frozensets
        Discovered resource groups.
    
    Notes
    -----
    Using this function relies on an external tool, OSLOM [1]_. This 
    function provides merely the interface between OrgMiner and the 
    external tool.

    See Also
    --------
    orgminer.SocialNetworkMiner.joint_activities
    scipy.spatial.distance

    References
    ----------
    .. [1] Lancichinetti, A., Radicchi, F., Ramasco, J. J., & Fortunato,
       S. (2011). Finding statistically significant communities in
       networks. *PloS one*, 6(4), e18961.
       `<https://doi.org/10.1371/journal.pone.0018961>`_
    """
    print('Applying graph/network-based local expansion method ' + 
        '(overlapping community detection using OSLOM):')
    # build network from profiles
    from orgminer.SocialNetworkMiner.joint_activities import distance
    sn = distance(profiles, metric=metric, convert=True)

    # step 0. Relabel nodes
    sn, rev_node_relabel_mapping = _relabel_nodes_integers(sn)

    # step 1. Distinguish the isolated nodes
    from networkx import isolates
    original_isolates = list(isolates(sn))
    if len(original_isolates) > 0:
        warn('There exist {} isolated nodes in the network.'.format(
            len(original_isolates)), RuntimeWarning)

    # step 2. Export network as edgelist
    from networkx import write_weighted_edgelist
    with open('tmp_sn.edgelist', 'wb') as f:
        write_weighted_edgelist(sn, f)
    print('Network exported to "tmp_sn.edgelist".')
    print('Use the external tool OSLOM to discover communities:')

    # step 3. Run community detection applying OSLOM
    print('Path to the community detection result file as input: ', end='')
    fn_communities = input()

    # step 4. Derive organizational groups from the detection results
    from collections import defaultdict
    groups = defaultdict(set)
    cnt = -1
    with open(fn_communities, 'r') as f:
        for line in f:
            if not line.startswith('#'): # for non-comment lines
                cnt += 1
                for label in line.split():
                    # restore labels 
                    groups[cnt].add(rev_node_relabel_mapping[int(label)])
    print('Detected communities imported from "{}":'.format(fn_communities))

    for i, iso_node in enumerate(original_isolates):
        groups['ISOLATE #{}'.format(i)].add(rev_node_relabel_mapping[iso_node])
    return [frozenset(g) for g in groups.values()]


def agent_copra(profiles, metric='euclidean'):
    """Apply an agent-based technique [1]_ for detecting communities 
    and thus to discover organizational groups.

    Parameters
    ----------
    profiles : DataFrame
        Constructed resource profiles.
    metric : str, optional, default 'euclidean'
        Choice of metrics for measuring the distance while calculating 
        distance. Defaults to ``'euclidean'``, meaning that euclidean
        distance is used for measuring distance.

    Returns
    -------
    list of frozensets
        Discovered resource groups.
    
    Notes
    -----
    Using this function relies on an external tool, COPRA [1]_. This 
    function provides merely the interface between OrgMiner and the 
    external tool.
    
    See Also
    --------
    orgminer.SocialNetworkMiner.joint_activities
    scipy.spatial.distance
    agent_slpa

    References
    ----------
    .. [1] Gregory, S. (2010). Finding overlapping communities in
       networks by label propagation. *New Journal of Physics*, 12(10),
       103018. `<https://doi.org/10.1088/1367-2630/12/10/103018>`_
    """
    print('Applying graph/network-based agent-based method ' + 
        '(overlapping community detection using COPRA):')
    # build network from profiles
    from orgminer.SocialNetworkMiner.joint_activities import distance
    sn = distance(profiles, metric=metric, convert=True)

    # step 0. Relabel nodes
    sn, rev_node_relabel_mapping = _relabel_nodes_integers(sn)

    # step 1. Distinguish the isolated nodes
    from networkx import isolates
    original_isolates = list(isolates(sn))
    if len(original_isolates) > 0:
        warn('There exist {} isolated nodes in the network.'.format(
            len(original_isolates)), RuntimeWarning)

    # step 2. Export network as edgelist
    from networkx import write_weighted_edgelist
    with open('tmp_sn.edgelist', 'wb') as f:
        write_weighted_edgelist(sn, f)
    print('Network exported to "tmp_sn.edgelist".')
    print('Use the external tool COPRA to discover communities:')

    # step 3. Run community detection applying OSLOM
    print('Path to the community detection result file as input: ', end='')
    fn_communities = input()

    # step 4. Derive organizational groups from the detection results
    from collections import defaultdict
    groups = defaultdict(set)
    cnt = -1
    with open(fn_communities, 'r') as f:
        for line in f:
            cnt += 1
            for label in line.split():
                # restore labels 
                groups[cnt].add(rev_node_relabel_mapping[int(label)])
    print('Detected communities imported from "{}":'.format(fn_communities))

    for i, iso_node in enumerate(original_isolates):
        groups['ISOLATE #{}'.format(i)].add(rev_node_relabel_mapping[iso_node])
    return [frozenset(g) for g in groups.values()]


def agent_slpa(profiles, metric='euclidean'):
    """Apply an agent-based technique [1]_ for detecting communities 
    and thus to discover organizational groups.

    Parameters
    ----------
    profiles : DataFrame
        Constructed resource profiles.
    metric : str, optional, default 'euclidean'
        Choice of metrics for measuring the distance while calculating 
        distance. Defaults to ``'euclidean'``, meaning that euclidean
        distance is used for measuring distance.

    Returns
    -------
    ogs : list of frozensets
        Discovered resource groups.
    
    Notes
    -----
    Using this function relies on an external tool, GANXiSw [1]_. This 
    function provides merely the interface between OrgMiner and the 
    external tool.

    See Also
    --------
    orgminer.SocialNetworkMiner.joint_activities
    scipy.spatial.distance
    agent_copra

    References
    ----------
    .. [1] Xie, J., Szymanski, B. K., & Liu, X. (2011). Slpa: Uncovering
       overlapping communities in social networks via a speaker-listener
       interaction dynamic process. In *Proceedings of the 2011 IEEE 11th
       International Conference on data mining workshops*, pp. 344-349.
       IEEE. `<https://doi.org/10.1109/ICDMW.2011.154>`_
    """
    print('Applying graph/network-based agent-based method ' + 
        '(overlapping community detection using SLPA):')
    # build network from profiles
    from orgminer.SocialNetworkMiner.joint_activities import distance
    sn = distance(profiles, metric=metric, convert=True)

    # step 0. Relabel nodes
    sn, rev_node_relabel_mapping = _relabel_nodes_integers(sn)

    # step 1. Distinguish the isolated nodes
    from networkx import isolates
    original_isolates = list(isolates(sn))
    if len(original_isolates) > 0:
        warn('There exist {} isolated nodes in the network.'.format(
            len(original_isolates)), RuntimeWarning)

    # step 2. Export network as edgelist
    from networkx import write_weighted_edgelist
    with open('tmp_sn.edgelist', 'wb') as f:
        write_weighted_edgelist(sn, f)
    print('Network exported to "tmp_sn.edgelist".')
    print('Use the external tool GANXiS (aka SLPA) to discover communities:')

    # step 3. Run community detection applying OSLOM
    print('Path to the output directory as input: ', end='')
    dirn_communities = input()

    # step 4. Derive organizational groups from the detection results
    # Note: since GANXiSw produces a set of results varied by the prob. "r"
    from collections import defaultdict
    best_fn = None
    best_eq = float('-inf')
    fn_cnt = 0
    solution = None
    from os import listdir, path
    for fn in listdir(dirn_communities):
        if fn.endswith('.icpm'):
            fn_cnt += 1
            fn_communities = path.join(dirn_communities, fn)
            cnt = -1
            groups = defaultdict(set)
            with open(fn_communities, 'r') as f:
                for line in f:
                    cnt += 1
                    for label in line.split():
                        groups[cnt].add(int(label))
            for i, iso_node in enumerate(original_isolates):
                groups['ISOLATE #{}'.format(i)].add(iso_node)
            eq = _extended_modularity(sn, list(groups.values()))
            if eq > best_eq:
                best_fn = fn_communities
                best_eq = eq
                solution = groups

    print('Detected communities imported from directory "{}":'.format(
        dirn_communities))
    print('Best solution "{}" selected from {} candidates:'.format(
        best_fn, fn_cnt))
    print('{} organizational groups discovered.'.format(len(solution)))
    # restore labels
    ogs = list()
    for cover in solution.values():
        ogs.append(frozenset({rev_node_relabel_mapping[x] for x in cover}))
    return ogs




# -*- coding: utf-8 -*-

"""This module contains definitions of several classes that are used by 
the clustering-based methods for organizational model discovery.

See Also
--------
orgminer.OrganizationalModelMiner.clustering.overlap
"""
from warnings import warn

class MOC:
    """This class implements the method of Model-based Overlapping 
    Clustering (MOC) [1]_.

    Methods
    -------
    fit_predict(X)
        Fit and then predict labels for the input samples.
    score(X, M, A)
        Calculate the likelihood score of a solution from fitting the 
        samples

    Notes
    -----
    In the current implementation, Gaussian distributions are assumed for
    the latent categories of the data samples, thus squared Euclidean
    distance measure is used as the Bregman divergence.

    References
    ----------
    .. [1] Banerjee, A., Krumpelman, C., Ghosh, J., Basu, S., & Mooney,
       R. J. (2005). Model-based overlapping clustering. In
       *Proceedings of the eleventh ACM SIGKDD International Conference 
       on Knowledge Discovery in Data mining*, pp. 532-537. ACM.
       `<https://doi.org/10.1145/1081870.1081932>`_
    """

    def __init__(self, 
        n_components, tol=1e-6, n_init=1, max_iter=1000, M_init=None,
        is_disjoint=False):
        """Instantiate an MOC class instance.

        Parameters
        ----------
        n_components : int
            Number of components in the model.
        tol : float, optional, default 1e-6
            The convergence threshold.
        n_init : int, optional, default 1
            Number of initializations to perform. The best results 
            are kept. This parameter would be override `M_init` if 
            specified.
        max_iter : int, optional, default 1000
            Number of iterative alternating updates to run.
        M_init : array-like, shape (n_samples, n_components), optional, 
        default None
            User-provided initial membership matrix M (binary-valued). 
            If ``None``, random initialization is used.
        is_disjoint : bool, optional, default False
            A boolean flag indicating whether a disjoint result is 
            required.
        """
        self._n_components = n_components
        self._tol = tol
        self._n_init = n_init if M_init is None else 1
        self._max_iter = max_iter
        self._M_init = M_init
        self._is_disjoint = is_disjoint


    def _init_M(self, n_init, shape):
        """Initialize a list of random guesses of membership M.

        Parameters
        ----------
        n_init : int
            Number of iterations to be used.

        Returns
        -------
        l_rand_M : list
            The result of random initialization.
        """
        from numpy import zeros
        from numpy.random import randint, choice

        l_rand_M = list()
        visited = set() # pool for storing used seeds
        for init in range(n_init):
            # Random initialization (non-repeat)
            while True:
                M = zeros(shape)
                if self._is_disjoint:
                    for row in M:
                        row[randint(self._n_components)] = 1 # only 1
                else:
                    for row in M:
                        row[choice(range(shape[1]),
                            size=randint(1, shape[1] + 1),
                            replace=False)] = 1

                M.flags.writeable = False # set RO, entries immutable
                bM = M.data.tobytes()
                if bM not in visited:
                    # if seed unused before
                    visited.add(bM)
                    l_rand_M.append(M)
                    break
        return l_rand_M
    

    def fit_predict(self, X):
        """Fit and then predict labels for the input samples.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        best_m : array-like, shape (n_samples, n_components)
            The membership component in the model.
        """
        from numpy import array, dot, infty
        from numpy.linalg import pinv

        if self._M_init is not None:
            l_M = [self._M_init.copy()]
        else:
            l_M = self._init_M(self._n_init, shape=(len(X), self._n_components))
            
        def _e_m(M):
            # Start fitting with valid initialized value of M
            iteration = -1
            current_log_likelihood = None
            delta_log_likelihood = infty
            converged = False

            while not converged:
                iteration += 1
                if iteration >= self._max_iter:
                    warn('Convergence not reached within {} iterations'.format(
                        self._max_iter), RuntimeWarning)
                    return M

                #print('\n\tIteration {}:'.format(iteration), end=' ')
                prev_log_likelihood = current_log_likelihood

                # Alternate between updating M and A
                # update A: assume loss function squared Euclidean Distance
                pinv_M = pinv(M) 
                A = dot(pinv_M, X)

                # update M: for each row, apply appropriate search algorithms
                new_M = list()
                for i in range(X.shape[0]):
                    #m_best = self._enumerate(X[i,:], M[i,:], A)
                    m_best = self._dynamicm(X[i,:], M[i,:], A)
                    new_M.append(m_best)
                prev_M = M.copy()
                M = array(new_M)

                # Calculate new log-likelihood:
                current_log_likelihood = self.score(X, M, A)
                #print('score = {:.8f}'.format(current_log_likelihood), end='')

                if prev_log_likelihood is not None: # if not the initial run
                    delta_log_likelihood = (current_log_likelihood
                            - prev_log_likelihood)

                if delta_log_likelihood is not infty and delta_log_likelihood < self._tol:
                    converged = True
                    if delta_log_likelihood >= 0:
                        pass
                    else:
                        # current solution is worse (maximizing), use the last one
                        #print('\tDELTA < 0: Pick the last solution instead.')
                        current_log_likelihood = prev_log_likelihood
                        M = prev_M.copy()
                    #print('\nModel converged with ', end='')

            # check if the solution is valid
            is_valid = True
            for j in range(self._n_components):
                if not M[:,j].any(): # invalid, an empty cluster found
                    is_valid = False
                    return M.copy(), float('-nan') # return as NaN score
            if is_valid:
                #print('Final score =\t{:.8f}'.format(current_log_likelihood))
                return M.copy(), current_log_likelihood
       
        l_fitted_M = list(map(_e_m, l_M))
        best_M, best_score = max(l_fitted_M, key=lambda x: x[1])
        return best_M


    # Definition of the likelihood function: log p(X, M, A) to be maximized
    def score(self, X, M, A):
        """Calculate the likelihood score of a solution from fitting the 
        samples.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.
        M : array-like, shape (n_samples, n_components)
            The membership component in the model.
        A : array-like, shape (n_components, n_features)
            The parameter component in the model.
        
        Returns
        -------
        float
            The result score.

        See Also
        --------
        fit_predict
        """
        from numpy import dot, log, power
        #print('Calculating the score: ', end='')
        # calculate alpha_ih
        score_alpha = 0.0
        for h in range(M.shape[1]): # n_components
            M_h = M[:,h]
            pie_h = (1.0 / X.shape[0]) * len(M_h[M_h == True])
            for i in range(M.shape[0]): # n_samples
                # Bernoulli
                alpha_ih = (power(pie_h, M[i,h])
                        * power((1 - pie_h), (1 - M[i,h])))
                score_alpha += log(alpha_ih)

        # calculate the Bregman divergence
        # currently Squared Euclidean Distance
        score_divergence = 0.0
        from scipy.spatial.distance import sqeuclidean
        MA = dot(M, A)
        for i in range(X.shape[0]): # n_components
            for j in range(X.shape[1]): # n_features
                score_divergence += sqeuclidean(X[i,j], MA[i,j])

        score = score_alpha - score_divergence
        #print('{:.4f}'.format(score))
        return score
    

    '''
    # search over all possible settings in the naive way
    def _enumerate(self, x, m0, A):
        from numpy import array, dot
        from scipy.spatial.distance import sqeuclidean
        from itertools import combinations

        L0 = sqeuclidean(x, dot(m0, A))
        Lmin = L0
        k = len(m0)
        m_best = m0

        # Exhaustively enumerate all possible settings sum_{i = 1 to k} C^i_k
        # start from 1 ensures that at least one scalar with value True
        for i in range(1, k + 1):
            for index in combinations(range(k), i):
                m = array([0] * k)
                m[[index]] = 1
                L = sqeuclidean(x, dot(m, A))
                if L < Lmin:
                    Lmin = L
                    m_best = m
        return m_best
    '''


    def _search_thread(self, x, A, m):
        from numpy import array, dot, infty, where
        from scipy.spatial.distance import sqeuclidean
        # generate the combinations as candidate positions
        L_min = sqeuclidean(x, dot(m, A))

        is_active = True
        while is_active: # search each "level"
            best_candidate = None
            # choose a best candidate to "turn on" from the rest
            for i in where(m == 0)[0]:
                candidate_m = m.copy()
                candidate_m[i] = 1 # try turning on 1 cluster
                candidate_L = sqeuclidean(
                    x, dot(candidate_m, A))
                if candidate_L < L_min:
                    # update if better than the current one
                    best_candidate = i
                    L_min = candidate_L
            if best_candidate is not None:
                is_active = True
                m[best_candidate] = 1
            else:
                # if no better frontier could be found, set inactive
                is_active = False # break
        return m, sqeuclidean(x, dot(m, A))


    # search on each separate threads with different initial setting
    # greedily proceed on the fly (idea similar to DP)
    def _dynamicm(self, x, m0, A):
        """A dynamic algorithm (using parallel threads) to search for a 
        "best" solution.

        Parameters
        ----------
        x : array-like, shape (n_features, )
            A row of the input samples.
        m0 : array-like, shape (n_components, )
            A row of the membership component in the model.
        A : array-like, shape (n_components, n_features)
            The parameter component in the model.
        
        Returns
        -------
        m_best : array-like, shape (n_components, )
            A "best" solution with respect to the given values.

        Notes
        -----
        The idea is similar to Dynamic Programming. The returned result 
        is prone to be sub-optimal. Refer to the paper for more 
        information.

        See Also
        --------
        fit_and_predict
        """
        from numpy import array, dot, infty, where
        from scipy.spatial.distance import sqeuclidean

        n_components = len(m0)
        L0 = sqeuclidean(x, dot(m0, A))
        m_best = m0

        if self._is_disjoint:
            for h in range(n_components):
                m = array([0] * n_components)
                m[h] = 1
                L = sqeuclidean(x, dot(m, A))
                if L < L0:
                    L0 = L
                    m_best = m
        else:
            from functools import partial
            from multiprocessing import Pool
            from operator import itemgetter

            separate_search_threads = list()
            for h in range(n_components): # init each search thread
                m = array([0] * n_components)
                m[h] = 1
                separate_search_threads.append(m)

            with Pool() as p:
                m_best = min(
                    p.map(
                        partial(self._search_thread, x, A),
                        separate_search_threads
                    ), 
                    key=itemgetter(1))[0]
        return m_best


class FCM:
    """This class implements the method of Fuzzy C-Means clustering [2]_.

    The implementation is based on `SciKit-Fuzzy 
    <http://pythonhosted.org/scikit-fuzzy/>`_.

    Methods
    -------
    fit_predict(X) : Fit and then predict labels for the input samples.

    References
    ----------
    .. [2] Tan, P. N., Steinbach, M., Karpatne, A., & Kumar, V. (2018).
       *Introduction to Data Mining*.
    """

    def __init__(self,
        n_components, tol=1e-6, p=2, n_init=1, max_iter=1000,
        means_init=None):
        """Instantiate an FCM class instance.

        Parameters
        ----------
        n_components : int
            Number of clusters expected.
        tol : float, optional, default 1e-6
            The convergence threshold.
        p : float, optional, default 2
            The exponentiation value. When p = 1, fuzzy c-means reduces
            to K-means algorithm; when p is set to larger values, the 
            partitioning becomes fuzzier (approaching global centroid).
        n_init : int, optional, default 1
            Number of initializations to perform. The best results 
            are kept. This parameter would be override `means_init if 
            specified.
        max_iter : int, optional, default 1000
            Number of iterative alternating updates to run.
        means_init : array-like, shape (n_components, n_features), 
        optional, default None
            User-provided initial guess of centroids. If ``None``, random
            initialization is used and assigns random-valued weights for
            samples.
        """
        self._n_components = n_components
        self._tol = tol
        self._p = p
        self._n_init = n_init if means_init is None else 1
        self._max_iter = max_iter
        self._means_init = means_init


    def _init_w(self, n_init, shape):
        """Initialize a list of random guesses of fuzzy pseudo partition 
        w.

        Parameters
        ----------
        n_init : int
            Number of iterations to be used.

        Returns
        -------
        l_rand_w : list
            The result of random initialization.
        """
        from numpy import array
        from numpy.random import randint, choice

        l_rand_w = list()
        for init in range(n_init):
            # random init, constraint: row sum = 1.0
            w = list()
            for i in range(shape[0]):
                weights = randint(1, 10, shape[1]) 
                weights = weights / sum(weights)
                w.append(weights)
            l_rand_w.append(array(w))
        return l_rand_w


    def fit_predict(self, X):
        """Fit and then predict labels for the input samples.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        best_w : array-like, shape (n_samples, n_components)
            The weight values.
        """
        from numpy import array, infty, zeros, sum, power, dot, amax
        from scipy.spatial.distance import euclidean as dist
        from skfuzzy.cluster import cmeans

        best_w = None
        if self._means_init is not None:
            # if seed centroids given, compute initial fuzzy pseudo partition
            w = zeros((len(X), self._n_components))
            exp = 1 / (self._p - 1)
            for i in range(len(X)):
                l_sqd_xi_c = [power(dist(X[i,:], self._means_init[q,:]), 2)
                        for q in range(self._n_components)]
                if 0.0 in l_sqd_xi_c:
                    # special case: current point is one of the centroids
                    cntr_cluster_ix = l_sqd_xi_c.index(0.0)
                    w[i,cntr_cluster_ix] = 1.0 # leave others 0
                else:
                    for j in range(self._n_components):
                        sqd_xi_cj = power(dist(X[i,:], self._means_init[j,:]), 2)
                        w[i,j] = (
                                power((1 / sqd_xi_cj), exp)
                                / sum([power(
                                    (1 / power(dist(X[i,:], self._means_init[q,:]), 2)), 
                                    exp) for q in range(self._n_components)]))
            l_w = [w.copy()]
        else:
            l_w = self._init_w(self._n_init, shape=(len(X), self._n_components))

        def _e_m(w):
            _, w, w0, _, sse, _, _ = cmeans(data=X.T,
                    c=self._n_components, m=self._p, error=self._tol, 
                    maxiter=self._max_iter, init=w.T)
             
            # check if the solution is valid
            is_valid = True
            for j in range(self._n_components):
                if not w[:,j].any():
                    is_valid = False
                    break

            if is_valid:
                #print('Final SSE =\t{:.8f}'.format(sse[-1]))
                return w.copy().T, sse[-1]
        
        l_fitted_w = list(map(_e_m, l_w))
        best_w, best_sse = max(l_fitted_w, key=lambda x: x[1])
        return best_w




# -*- coding: utf-8 -*-

"""This module provides a set of intrinsic evaluation measures for the 
latent clusters (grouping of resources in the organizational models). 

Notes
-----
To conduct the validation of clusters, the expected inputs of the 
included functions should be both the clusters and the resource feature 
table (profile matrix), rather than an organizational model instance.

Under the situation where the clustering was derived with no feature 
table was used but rather from a graph-like structure, the modularity 
measure should be used. The inputs should be the clusters and the
original graph-like structure from which the clustering was derived.
"""
from warnings import warn

# TODO: is it appropriate using silhouette score for overlapped clusters?
def silhouette_score(clu, X, metric='euclidean'):
    """Calculate the silhouette scores of the latent clustering.
    
    Parameters
    ----------
    clu : list of frozensets
        Resource groups in an organizational model.
    X : DataFrame
        The corresponding resource feature table used for deriving the 
        resource groups.
    metric : str, optional
        Distance measure used for deriving the resource groups. Refer to
        `scipy.spatial.distance <https://docs.scipy.org/doc/scipy/
        reference/spatial.distance.html>`_.

    Returns
    -------
    scores : dict
        The result silhouette score (of each resource, respectively).

    Notes
    -----
    This is merely a wrap-up of function
    `sklearn.metrics.silhouette_samples
    <https://scikit-learn.org/stable/modules/generated/sklearn.metrics.silhouette_samples.html>`_.
    """
    from sklearn.metrics import silhouette_samples
    resources = list(X.index)
    labels = [-1] * len(resources)
    for ig, g in enumerate(clu):
        for r in g:
            labels[resources.index(r)] = ig

    scores_samples = silhouette_samples(X.values, labels, metric=metric)
    scores = dict()
    for ir, r in enumerate(resources):
        scores[r] = scores_samples[ir]
    return scores


def variance_explained_score(clu, X):
    """Calculate the degree of variance explained by the latent 
    clustering, i.e. Calinski-Harabasz score.
    
    Parameters
    ----------
    clu : list of frozensets
        Resource groups in an organizational model.
    X : DataFrame
        The corresponding resource feature table used for deriving the 
        resource groups.

    Returns
    -------
    float
        The result score representing the degree of variance explained.

    Notes
    -----
    This is merely a wrap-up of function
    `sklearn.metrics.calinski_harabasz_score
    <https://scikit-learn.org/stable/modules/generated/sklearn.metrics.calinski_harabasz_score.html>`_.
    """
    from sklearn.metrics import calinski_harabasz_score 
    resources = list(X.index)
    labels = [-1] * len(resources)
    for ig, g in enumerate(clu):
        for r in g:
            labels[resources.index(r)] = ig
    return calinski_harabasz_score(X.values, labels)


def variance_explained_percentage(clu, X):
    """Calculate the percentage of variance explained by the latent 
    clustering.
    
    Parameters
    ----------
    clu : list of frozensets
        Resource groups in an organizational model.
    X : DataFrame
        The corresponding resource feature table used for deriving the 
        resource groups.

    Returns
    -------
    float
        The result percentage of variance explained (in range 0-100).
    """
    var_between = _variance_between_cluster(clu, X)
    var_within = _variance_within_cluster(clu, X)
    return 100 * var_between / (var_between + var_within)


def _variance_within_cluster(clu, X):
    """Calculate the variance for objects within the same clusters.
    
    Parameters
    ----------
    clu : list of frozensets
        Resource groups in an organizational model.
    X : DataFrame
        The corresponding resource feature table used for deriving the 
        resource groups.

    Returns
    -------
    var_between : float
        The result within-cluster variance.
    """
    from numpy import mean, sum
    var_within = 0

    for ig, g in enumerate(clu):
        g_mean = mean(X.loc[g].values, axis=0)
        var_within += sum((X.loc[g].values - g_mean) ** 2) # W_k
    var_within /= (len(X) - len(clu)) # N - k
    return var_within


def _variance_between_cluster(clu, X):
    """Calculate the variance for objects between different clusters.
    
    Parameters
    ----------
    clu : list of frozensets
        Resource groups in an organizational model.
    X : DataFrame
        The corresponding resource feature table used for deriving the 
        resource groups.

    Returns
    -------
    var_between : float
        The result between-cluster variance.
    """
    from numpy import mean, sum
    var_between = 0
    samples_mean = mean(X.values, axis=0)

    for ig, g in enumerate(clu):
        g_mean = mean(X.loc[g].values, axis=0)
        var_between += len(g) * sum((g_mean - samples_mean) ** 2) # B_k
    var_between /= (len(clu) - 1) # k - 1
    return var_between


def modularity(clu, G, weight=None):
    """Calculate the modularity score [1]_ of the latent clustering.
    
    Parameters
    ----------
    clu : list of frozensets
        Resource groups in an organizational model.
    G : NetworkX Graph or DiGraph
        The corresponding resource feature table used for deriving the 
        resource groups.
    weight : str, optional
        The name of the key defining the weight value of edges in `G`.

    Returns
    -------
    q : float
        The result modularity score.

    References
    ----------
    .. [1] "Graph Clustering Methods", Equation (11.39), Sect.
       11.3.3 in book, J. Han, J. Pei, M. Kamber. (2011). *Data Mining:
       Concepts and Techniques*.
    """
    from networkx import is_directed, restricted_view
    q = 0.0
    if is_directed(G):
        """Note: Casting from DiGraph to Graph needs to be configured 
        manually otherwise NetworkX would approach this in an arbitrary 
        fashion.
        """
        from itertools import combinations
        undirected_edge_list = list()
        # consider only pairwise links
        for pair in combinations(G.nodes, r=2):
            if (G.has_edge(pair[0], pair[1]) and 
                G.has_edge(pair[1], pair[0])):
                undirected_edge_wt = 0.5 * (
                    G[pair[0]][pair[1]]['weight'] +
                    G[pair[1]][pair[0]]['weight'])
                undirected_edge_list.append(
                    (pair[0], pair[1], {'weight': undirected_edge_wt}))
            else:
                pass
        G.clear()
        del G
        from networkx import Graph
        G = Graph()
        G.add_edges_from(undirected_edge_list)
        del undirected_edge_list[:]
        warn('DiGraph casted to Graph.', RuntimeWarning)
    else:
        pass

    sum_total_wt = sum([wt for (u, v, wt) in G.edges.data('weight')])
    for g in clu:
        sub_g = restricted_view(G, nodes=list(g), edges=[])
        if weight is None:
            sum_within_cluster_edges = len(list(sub_g.edges)) # l_i
        else:
            sum_within_cluster_edges = sum([wt
                for (u, v, wt) in sub_g.edges.data(weight)]) # l_i
        sum_within_cluster_degree = sum([deg
            for node, deg in sub_g.degree(weight=weight)]) # d_i
        q += (
            (sum_within_cluster_edges / sum_total_wt) - 
            (sum_within_cluster_degree / (2 * sum_total_wt)) ** 2)
    return q




# -*- coding: utf-8 -*-

"""This module provides several simple execution mode learning 
approaches based on external information other than the log data, namely

    - TraceClusteringCTMiner (CT only)
    - TraceClusteringFullMiner (CT & AT & TT)

"""
from .direct_groupby import CTonlyMiner, ATTTMiner

class TraceClusteringCTMiner(CTonlyMiner):
    """Informed by the result of applying trace clustering, each 
    variant of cases is taken as a case type.

    See Also
    --------
    orgminer.ExecutionModeMiner.direct_groupby.CTonlyMiner
    """

    def __init__(self, el, fn_partition):
        self._build_ctypes(el, fn_partition)
        CTonlyMiner._build_atypes(self, el)
        CTonlyMiner._build_ttypes(self, el)
        self._verify()


    def _build_ctypes(self, el, fn_partition):
        self._ctypes = dict()
        with open(fn_partition, 'r') as f_par:
            for line in f_par:
                case_id, cluster = (line.split('\t')[0].strip(), 
                    line.split('\t')[1].strip())
                self._ctypes[case_id] = 'CT.{}'.format(cluster)
        self.is_ctypes_verified = self._verify_partition(
            set(el['case_id']), self._ctypes)


class TraceClusteringFullMiner(TraceClusteringCTMiner, ATTTMiner):
    """Informed by the result of applying trace clustering, each 
    variant of cases is taken as a case type, each value of activity 
    (task) label is taken as an activity type, and each possible value 
    of a designated datetime unit is taken as a time type.

    See Also
    --------
    orgminer.ExecutionModeMiner.direct_groupby.FullMiner
    orgminer.ExecutionModeMiner.informed_groupby.TraceClusteringCTMiner
    """

    def __init__(self, el, 
        fn_partition, resolution, datetime_format='%Y/%m/%d %H:%M:%S.%f'):
        TraceClusteringCTMiner._build_ctypes(self, el, fn_partition)
        ATTTMiner._build_atypes(self, el)
        ATTTMiner._build_ttypes(self, el, resolution, datetime_format)
        self._verify()




# -*- coding: utf-8 -*-

from collections import defaultdict

class BaseMiner:
    """This class should be used as an interface base class for 
    implementing any approach for learning execution modes.

    Parameters
    ----------
    el : DataFrame
        An event log from which the execution modes are learned.

    Attributes
    ----------
    is_ctypes_verified : bool
        A boolean flag indicating whether the case types are verified.
    is_atypes_verified : bool
        A boolean flag indicating whether the activity types are
        verified.
    is_ttypes_verified : bool
        A boolean flag indicating whether the time types are verified.

    Methods
    -------
    derive_resource_log(el)
        Derive a resource log given the original log, after the 
        execution modes have been discovered and verified (which is done 
        the moment when an object is instantiated).
    get_type_by_value(value)
        Query the built type given a value (of either an activity 
        label, a case id, or a timestamp).
    get_values_by_type(type_name)
        Query the original values (of activity labels, case ids, or 
        timestamps) given a type name.

    Notes
    -----
    The docstrings for this class and its methods provide the guidelines 
    for any child class that implements a specific way of learning 
    execution modes by overwriting the three methods, respectively:

        - `self._build_ctypes`
        - `self._build_atypes`
        - `self._build_ttypes`
    
    Any child class should inherit from class `BaseMiner` and enables 
    learning and storing the mappings:

    - from case ids to Case Types,

    .. math:: \\mathcal{C} \\rightarrow \\mathcal{CT}

    - from activity labels to Activity Types,

    .. math:: \\mathcal{A} \\rightarrow \\mathcal{AT}

    - from timestamps to Time Types,

    .. math:: \\mathcal{T} \\rightarrow \\mathcal{TT}

    These mappings should be built as python `dicts` of strings.

    Please note that, any approach inherited from class `BaseMiner` and 
    implementing some execution mode learning strategy must ensure that:

        - the derived types are disjoint, and
        - the derived types form a partitioning of the original values.

    For example, a learning approach must map each case captured in the 
    event log to one case type and one case type only.
    """

    _ctypes = None
    is_ctypes_verified = False

    _atypes = None
    is_atypes_verified = False

    _ttypes = None
    is_ttypes_verified = False

    def __init__(self, el, **kwargs):
        """Instantiate an instance that implements a way of execution
        mode learning.

        The constructor method should invoke the three methods for 
        building case types, activity types and time types, 
        respectively:

        - `self._build_ctypes`
        - `self._build_atypes`
        - `self._build_ttypes`

        The function signature may vary depending on the exact strategy 
        of learning execution modes.

        And invoke method `self._verify()` to ensure that the built 
        types are valid, i.e., a set of such types would form a 
        partitioning of the original corresponding values.

        Note that any child class is required to have its own 
        constructor method written as there may be different 
        requirements for the input arguments. Even if there is no
        additional requirements as in the base class, it is expected
        that constructor of the child class explicitly invokes this 
        method.

        Parameters
        ----------
        el : DataFrame
            An event log from which the execution modes are learned.

        Returns
        -------
        """
        self._build_ctypes(el)
        self._build_atypes(el)
        self._build_ttypes(el)
        self._verify()


    def _build_ctypes(self, el, **kwargs):
        """Mine the case types.

        Each type should be stored as a key-value pair where the key is 
        of string type and the value is a set of strings.

        By default, this dimension is neglected, i.e., no case types are
        built, or you may consider all cases belonging to the same dummy
        type marked by an empty python string, ``''``.

        Parameters
        ----------
        el : DataFrame
            An event log from which the execution modes are learned.

        Returns
        -------
        """
        self._ctypes = defaultdict(str) # defaults to ''
        self.is_ctypes_verified = True


    def _build_atypes(self, el, **kwargs):
        """Mine the activity types.

        Each type should be stored as a key-value pair where the key is 
        of string type and the value is a set of strings.

        By default, this dimension is neglected, i.e., no activity types
        are built, or you may consider all activity labels belonging to 
        the same dummy type marked by an empty python string, ``''``.

        Parameters
        ----------
        el : DataFrame
            An event log from which the execution modes are learned.

        Returns
        -------
        """
        self._atypes = defaultdict(str) # defaults to ''
        self.is_atypes_verified = True


    def _build_ttypes(self, el, **kwargs):
        """Mine the time types.

        Each type should be stored as a key-value pair where the key is 
        of string type and the value is a set of strings.

        By default, this dimension is neglected, i.e., no time types
        are built, or you may consider all timestamps belonging to 
        the same dummy type marked by an empty python string, ``''``.

        Parameters
        ----------
        el : DataFrame
            An event log from which the execution modes are learned.

        Returns
        -------
        """
        self._ttypes = defaultdict(str) # defaults to ''
        self.is_ttypes_verified = True


    def derive_resource_log(self, el):
        """Derive a resource log given the original log, after the 
        execution modes have been discovered and verified (which is done 
        the moment when an object is instantiated).

        Each resource event in the derived resource log is corresponded 
        with exactly an event in the source event log even when the 
        resource log is defined as a multiset.

        Note that, such resource events are required to contain resource
        information, i.e. events with no resource information in the 
        source event log will be discarded.

        Parameters
        ----------
        el : DataFrame
            An event log from which the execution modes are learned.

        Returns
        -------
        DataFrame
            The derived resource log as a pandas DataFrame.
        """
        rl = list()
        for event in el.itertuples(): # keep order
            # NOTE: only events with resource information are considered
            if event.resource is not None and event.resource != '':
                rl.append({
                    'resource': event.resource,
                    'case_type': self._ctypes[event.case_id],
                    'activity_type': self._atypes[event.activity],
                    'time_type': self._ttypes[event.timestamp]
                })

        from pandas import DataFrame
        rl = DataFrame(rl)
        print('Resource Log derived: ', end='')
        print('{} events mapped onto {} execution modes.\n'.format(
            len(el), 
            len(rl.drop_duplicates(
                subset=['case_type', 'activity_type', 'time_type'],
                inplace=False))
        ))
        return rl


    def to_file(self, f):
        """Export the constructed execution mode miner to an external 
        file using Python pickle.
        
        Parameters
        ----------
        f : File object
            A destination file to be written.

        Returns
        -------

        Raises
        ------
        RuntimeError
            If the constructed execution mode miner is not verified.
        """
        if (self.is_ctypes_verified and self.is_atypes_verified and
            self.is_ttypes_verified):
            obj = {
                '_ctypes': self._ctypes,
                'is_ctypes_verified': self.is_ctypes_verified,
                '_atypes': self._atypes,
                'is_atypes_verified': self.is_atypes_verified,
                '_ttypes': self._ttypes,
                'is_ttypes_verified': self.is_ttypes_verified
            }
            from json import dump
            dump(obj, f)
        else:
            raise RuntimeError('Unable to verify execution modes.')


    @classmethod
    def from_file(cls, f):
        """Import a constructed execution mode miner to an external file 
        using Python pickle.

        Parameters
        ----------
        f: File object
            A sourced file to be read from.

        Returns
        -------
        (Corresponding miner class object)
            The imported constructed execution mode miner.

        Raises
        ------
        RuntimeError
            If the imported execution mode miner is not verified.
        """
        from json import load
        obj = load(f)
        if (obj['is_ctypes_verified'] and obj['is_atypes_verified'] and
            obj['is_ttypes_verified']):
            ret = cls(el=None)
            ret.is_ctypes_verified = ret.is_atypes_verified = \
                ret.is_ttypes_verified = True
            ret._ctypes = obj['_ctypes'] if len(obj['_ctypes']) > 0 else\
                defaultdict(str)
            ret._atypes = obj['_atypes'] if len(obj['_atypes']) > 0 else\
                defaultdict(str)
            ret._ttypes = obj['_ttypes'] if len(obj['_ttypes']) > 0 else\
                defaultdict(str)

            ret._verify()
            return ret
        else:
            raise RuntimeError('Unable to verify execution modes.')


    def _verify_partition(self, whole_set, partitioning):
        """A helper function that is used for verifying if the keys in a 
        given partitioning (as a dict) is indeed a partitioning of 
        values in a given set.

        Parameters
        ----------
        whole_set : set

        partitioning : dict

        Returns
        -------
        bool
            The result of verification.
        """
        is_disjoint = True # given the internal data structure as a dict
        is_union = set(partitioning.keys()) == whole_set
        return is_disjoint and is_union


    def _verify(self):
        """Verify if the built execution modes are valid and output
        prompting information.

        Parameters
        ----------

        Returns
        -------

        Raises
        ------
        RuntimeError
            If the built execution modes cannot be validated.
        """
        if (self.is_ctypes_verified and self.is_atypes_verified and
                self.is_ttypes_verified):
            print('-' * 80)
            print('Count of Types in the current {}:'.format(
                self.__class__.__name__))
            print('Number of C Types:\t\t{}'.format(
                len(set(self._ctypes.values())) 
                if self._ctypes is not None else 'n/a (1)'))
            print('Number of A Types:\t\t{}'.format(
                len(set(self._atypes.values()))
                if self._atypes is not None else 'n/a (1)'))
            print('Number of T Types:\t\t{}'.format(
                len(set(self._ttypes.values()))
                if self._ttypes is not None else 'n/a (1)'))
            print('-' * 80)
        else:
            print('C Types:\t{}'.format('VERIFIED' if self.is_ctypes_verified
                else 'INVALID'))
            print('A Types:\t{}'.format('VERIFIED' if self.is_atypes_verified
                else 'INVALID'))
            print('T Types:\t{}'.format('VERIFIED' if self.is_ttypes_verified
                else 'INVALID'))
            raise RuntimeError('Unable to verify execution modes.')


    # TODO:
    # from original activity labels/case ids/timestamps to types
    def get_type_by_value(self, value):
        """Query the built type given a value (of either an activity 
        label, a case id, or a timestamp).

        Parameters
        ----------
        value : str
            The given value to be queried.

        Returns
        -------
        type_name : str
            The corresponding type name.
        """
        raise NotImplementedError


    # from types to the originals
    def get_values_by_type(self, type_name):
        """Query the original values (of activity labels, case ids, or 
        timestamps) given a type name.

        Parameters
        ----------
        type_name : str
            The given type name to be queried.

        Returns
        -------
        values : list of str
            The corresponding values.

        Raises
        ------
        ValueError
            If the given type name being queried is invalid.
        """
        if type_name.startswith('CT'):
            return list(k for k, v in self._ctypes.items()
                if v == type_name)

        if type_name.startswith('AT'):
            return list(k for k, v in self._atypes.items()
                if v == type_name)

        if type_name.startswith('TT'):
            return list(k for k, v in self._ttypes.items()
                if v == type_name)

        raise ValueError('Invalid value for parameter `{}`: {}'.format(
            'type_name', type_name))
    



# -*- coding: utf-8 -*-

"""This module contains the implementation of the global conformance 
checking measures.
"""
from deprecated import deprecated
from warnings import warn

def _is_conformed_event(event, om):
    m = (event.case_type, event.activity_type, event.time_type)
    cand_groups = om.find_candidate_groups(m)
    for g in cand_groups:
        if event.resource in g:
            return True
    return False


def _is_allowed_event(event, om):
    m = (event.case_type, event.activity_type, event.time_type)
    cand_groups = om.find_candidate_groups(m)
    return len(cand_groups) > 0


def fitness(rl, om):
    """Calculate the fitness of an organizational model against a given
    resource log.

    Parameters
    ----------
    rl : DataFrame
        A resource log.
    om : OrganizationalModel
        The discovered organizational model.

    Returns
    -------
    float
        The result fitness value.

    Notes
    -----
    A resource log instead of an event log is used here, however this 
    is valid since a resource log is (implicitly) one-to-one 
    corresponded with an event log.
    """
    conformed_events = rl[rl.apply(
        lambda e: _is_conformed_event(e, om), axis=1)]
    n_conformed_events = len(conformed_events) # "|E_conf|"
    n_events = len(rl) # "|E_res|"
    return n_conformed_events / n_events


@deprecated(reason='This definition is neither being nor intended to be used.')
# fitness (by resource events)
def fitness_re(rl, om):
    conformed_events = rl[rl.apply(
        lambda e: _is_conformed_event(e, om), axis=1)]
    # "|RE_conf|"
    n_conformed_res_events = len(conformed_events.drop_duplicates())
    n_actual_res_events = len(rl.drop_duplicates()) # "|RE|"
    return n_conformed_res_events / n_actual_res_events


@deprecated(reason='This definition is not being used.')
# Wil's precision
def rc_measure(rl, om):
    """Calculate the precision of an organizational model against a 
    given event log. 
    
    Calculate model precision considering only "fitting" (conformed) 
    events.

    Parameters
    ----------
    rl : DataFrame
        A resource log.
    om : OrganizationalModel
        The discovered organizational model.

    Returns
    -------
    float
        The result precision value.

    Notes
    -----
    A resource log instead of an event log is used here, however this 
    is valid since a resource log is (implicitly) one-to-one 
    corresponded with an event log.
    """
    conformed_events = rl[rl.apply(
        lambda e: _is_conformed_event(e, om), axis=1)]
    n_conformed_events = len(conformed_events) # "|E_conf|"
    
    l_n_cand_e = list() # list of "cand(e)"
    cand_E = set()
    for event in conformed_events.itertuples():
        m = (event.case_type, event.activity_type, event.time_type)
        cand_groups = om.find_candidate_groups(m)

        cand_e = frozenset.union(*cand_groups) # cand(e)
        l_n_cand_e.append(len(cand_e)) # "|cand(e)|"
        cand_E.update(cand_e)

    n_cand_E = len(cand_E) # "|cand(E)|"

    if n_cand_E == 0:
        warn('No candidate resource.', RuntimeWarning)
        return float('nan')
    elif n_cand_E == 1:
        warn('Overall number of candidate is 1.', RuntimeWarning)
        return 1.0
    else:
        rc_sum = sum([(n_cand_E - n_cand_e) / (n_cand_E - 1) 
            for n_cand_e in l_n_cand_e])
        return rc_sum / n_conformed_events


@deprecated(reason='This definition is not being used.')
# precision (by resource events, not counting frequencies)
def precision_re_nofreq(rl, om):
    conformed_events = rl[rl.apply(
        lambda e: _is_conformed_event(e, om), axis=1)]
    # "|RE_conf|"
    n_conformed_res_events = len(conformed_events.drop_duplicates())
    # count of all possible resource events allowed by the model
    n_allowed_res_events = sum(len(om.find_execution_modes(r))
        for r in om.resources)
    return n_conformed_res_events / n_allowed_res_events


@deprecated(reason='This definition is not being used.')
# precision (by resource events, counting frequencies)
def precision_re_freq(rl, om):
    conformed_events = rl[rl.apply(
        lambda e: _is_conformed_event(e, om), axis=1)]
    n_conformed_events = len(conformed_events)

    # RE_conf
    conformed_res_events = set(
        (re.resource, re.case_type, re.activity_type, re.time_type)
        for re in conformed_events.drop_duplicates().itertuples())

    mode_occurrence = rl.groupby([
        'case_type', 'activity_type', 'time_type']).size().to_dict()

    F_conformed_res_events = 0.0
    F_allowed_res_events = 0.0
    F_model_false_res_events = 0.0
    for r in om.resources:
        for mode in om.find_execution_modes(r):
            res_event = (r, mode[0], mode[1], mode[2])
            if res_event in conformed_res_events:
                F_conformed_res_events += mode_occurrence[mode]
            else:
                F_model_false_res_events += mode_occurrence[mode]
            F_allowed_res_events += mode_occurrence[mode]

    return F_conformed_res_events / F_allowed_res_events


def precision(rl, om):
    """Calculate the precision of an organizational model against a 
    given event log. 
    
    Calculate model precision considering "allowed" events.

    Parameters
    ----------
    rl : DataFrame
        A resource log.
    om : OrganizationalModel
        The discovered organizational model.

    Returns
    -------
    float
        The result precision value.

    Notes
    -----
    A resource log instead of an event log is used here, however this 
    is valid since a resource log is (implicitly) one-to-one 
    corresponded with an event log.
    """
    cand_E = set()

    for event in rl.itertuples():
        if _is_allowed_event(event, om):
            m = (event.case_type, event.activity_type, event.time_type)
            cand_groups = om.find_candidate_groups(m)

            cand_e = frozenset.union(*cand_groups) # cand(e)
            cand_E.update(cand_e)
    n_cand_E = len(cand_E)

    if n_cand_E == 0:
        warn('No candidate resource.', RuntimeWarning)
        return float('nan')
    else:
        n_allowed_events = 0
        precision = 0.0

        for event in rl.itertuples():
            if _is_allowed_event(event, om):
                n_allowed_events += 1

                m = (event.case_type, event.activity_type, event.time_type)
                cand_groups = om.find_candidate_groups(m)
                cand_e = frozenset.union(*cand_groups) # cand(e)
                n_cand_e = len(cand_e)

                if _is_conformed_event(event, om):
                    # give reward
                    precision += (n_cand_E + 1 - n_cand_e) / n_cand_E
                else:
                    # NOTE: no extra penalty is given
                    precision += 0.0

        precision *= 1 / n_allowed_events
        return precision


def f1_score(rl, om):
    fitness_score = fitness(rl, om)
    precision_score = precision(rl, om)
    return ((2 * fitness_score * precision_score) 
            / (fitness_score + precision_score))




# -*- coding: utf-8 -*-

"""This module provides several simple execution mode learning 
approaches, namely

    - ATonlyMiner (AT only)
    - CTonlyMiner (CT only)
    - ATCTMiner (CT & AT)
    - ATTTMiner (AT & TT)
    - FullMiner (CT & AT & TT)

"""
from .base import BaseMiner

class ATonlyMiner(BaseMiner):
    """Each value of activity (task) label is taken as an activity type.
    """

    def __init__(self, el):
        BaseMiner.__init__(self, el)


    def _build_atypes(self, el):
        self._atypes = dict()
        for activity_label in set(el['activity']):
            self._atypes[activity_label] = 'AT.{}'.format(activity_label)

        self.is_atypes_verified = self._verify_partition(
            set(el['activity']), self._atypes)


class CTonlyMiner(BaseMiner):
    """Each value of a selected case-level attribute is taken as a case
    type.

    Raises
    ------
    ValueError
        If the specified case-level attribute does not qualify, i.e., 
        events from the same case record more than one values for this 
        attribute. Case id of the first problematic case will be 
        reported.

    """
    
    def __init__(self, el, case_attr_name):
        self._build_ctypes(el, case_attr_name)
        BaseMiner._build_atypes(self, el)
        BaseMiner._build_ttypes(self, el)
        self._verify()


    def _build_ctypes(self, el, case_attr_name):
        self._ctypes = dict()
        for v, events in el.groupby(case_attr_name): # sorted by default
            for case_id in set(events['case_id']):
                if case_id in self._ctypes:
                    raise ValueError('Inappropriate selection for ' +
                        '`case_attr_name`: check case id {}.'.format(case_id))
                else:
                    self._ctypes[case_id] = 'CT.{}'.format(v)

        self.is_ctypes_verified = self._verify_partition(
            set(el['case_id']), self._ctypes)


class ATCTMiner(ATonlyMiner, CTonlyMiner):
    """Each value of activity (task) label is taken as an activity type, 
    and each value of a selected case-level attribute is taken as a case 
    type.
    """

    def __init__(self, el, case_attr_name):
        CTonlyMiner._build_ctypes(self, el, case_attr_name)
        ATonlyMiner._build_atypes(self, el)
        ATonlyMiner._build_ttypes(self, el)
        self._verify()


class ATTTMiner(ATonlyMiner):
    """Each value of activity (task) label is taken as an activity type, 
    and each possible value of a designated datetime unit is taken as a 
    time type.
    """

    def __init__(self, el, resolution, datetime_format='%Y/%m/%d %H:%M:%S.%f'):
        ATonlyMiner._build_ctypes(self, el)
        ATonlyMiner._build_atypes(self, el)
        self._build_ttypes(el, resolution, datetime_format)
        self._verify()


    def _build_ttypes(self, el, resolution, datetime_format):
        self._ttypes = dict()
        from datetime import datetime
        if resolution in ['weekday', 'isoweekday']:
            # special case since (iso)weekday is a function of datetime
            # parse as 'isoweekday', i.e. Monday - 1, ..., Sunday - 7
            getter = lambda datetimeobj: datetimeobj.isoweekday()
        else:
            # otherwise access as read-only attribute
            from operator import attrgetter
            getter = attrgetter(resolution)

        for event in el.itertuples():
            dt = datetime.strptime(event.timestamp, datetime_format)
            self._ttypes[event.timestamp] = 'TT.{}'.format(getter(dt))

        self.is_ttypes_verified = self._verify_partition(
            set(el['timestamp']), self._ttypes)


class FullMiner(CTonlyMiner, ATTTMiner):
    """Each value of activity (task) label is taken as an activity type, 
    each value of a selected case-level attribute is taken as a case 
    type, and each possible value of a designated datetime unit is taken 
    as a time type.
    """

    def __init__(self, el, 
        case_attr_name, resolution, datetime_format='%Y/%m/%d %H:%M:%S.%f'):
        CTonlyMiner._build_ctypes(self, el, case_attr_name)
        ATTTMiner._build_atypes(self, el)
        ATTTMiner._build_ttypes(self, el, resolution, datetime_format)
        self._verify()




# -*- coding: utf-8 -*-

"""This module contains the implementation of overlapping organizational 
mining methods, based on the use of clustering techniques.
"""
def _gmm(profiles, n_groups, threshold, init='random', n_init=100): 
    """Apply the classic Gaussian Mixture Model (GMM) to discover 
    resource groups [1]_.

    Parameters
    ----------
    profiles : DataFrame
        Constructed resource profiles.
    n_groups : int
        Expected number of resource groups.
    threshold : float
        A given threshold value in range [0, 1.0] for producing
        determined clustering from the fuzzy clustering results from GMM.
    init : {'random', 'mja', 'ahc', 'kmeans', 'plain'}, optional, \
        default 'random'
        Options for deciding the strategy for initialization. Could be 
        one of the following:

            - ``'random'``, use random initialization.
            - ``'mja'``, use the mining method Metric based on Joint 
              Activities as initialization method.
            - ``'ahc'``, use the hierarchical method of Agglomerative
              Hierarchical Clustering as initialization method.
            - ``'kmeans'``, use the classic clustering algorithm kMeans 
              as initialization method.
            - ``'plain'``, simply put resources into groups by in the 
              order of their ids.

        Note that if an option other than ``'random'`` is specified, then
        the initialization is performed once only.
    n_init : int, optional, default 100
        Number of times of random initialization (if specified) performs
        before training the clustering model.

    Returns
    -------
    list of frozensets
        Discovered resource groups.
    
    Raises
    ------
    ValueError
        If the specified option for initialization is invalid.
    
    See Also
    --------
    sklearn.mixture.GaussianMixture
    orgminer.OrganizationalModelMiner.community.graph_partitioning.mja
    orgminer.OrganizationalModelMiner.clustering.hierarchical.ahc
    sklearn.cluster.KMeans

    References
    ----------
    .. [1] Yang, J., Ouyang, C., Pan, M., Yu, Y., & ter Hofstede, A.
       (2018). Finding the Liberos: Discover Organizational Models with
       Overlaps. In *Proceedings of the 16th International Conference on
       Business Process Management*, pp. 339-355. Springer, Cham.
       `<https://doi.org/10.1007/978-3-319-98648-7_20>`_
    """
    print('Applying overlapping clustering-based GMM:')
    # step 0. Perform specific initialization method (if given)
    if init in ['mja', 'ahc', 'kmeans', 'plain']:
        warm_start = True
        if init == 'mja':
            from orgminer.OrganizationalModelMiner.community.graph_partitioning\
                import _mja
            init_groups = _mja(profiles, n_groups)  
        elif init == 'ahc':
            from .hierarchical import _ahc
            init_groups, _ = _ahc(profiles, n_groups)
        elif init == 'kmeans':
            init_groups = list(set() for i in range(n_groups))
            from sklearn.cluster import KMeans
            labels = KMeans(n_clusters=n_groups, random_state=0).fit_predict(
                profiles)
            for i, r in enumerate(sorted(profiles.index)):
                init_groups[labels[i]].add(r)
        else: # init == 'plain'
            init_groups = list(set() for i in range(n_groups))
            for i, r in enumerate(sorted(profiles.index)):
                init_groups[i % n_groups].add(r)
        print('Initialization done using {}:'.format(init))
    elif init == 'random':
        warm_start = False
    else:
        raise ValueError('Invalid value for parameter `{}`: {}'.format(
            'init', init))

    # step 1. Train the model
    from sklearn.mixture import GaussianMixture
    if warm_start:
        from numpy import mean
        init_means = list()
        for g in init_groups:
            init_means.append(mean(profiles.loc[list(g)].values, axis=0))
        gmm_model = GaussianMixture(
            n_components=n_groups,
            covariance_type='tied',
            tol=1e-9,
            max_iter=1000,
            n_init=1,
            random_state=0,
            means_init=init_means).fit(profiles.values)
    else:
        gmm_model = GaussianMixture(
            n_components=n_groups,
            covariance_type='tied',
            tol=1e-9,
            max_iter=1000,
            n_init=n_init,
            init_params='random').fit(profiles.values)

    # step 2. Derive the clusters as the end result
    posterior_pr = gmm_model.predict_proba(profiles.values)
    from numpy import nonzero, median, unique, count_nonzero, amin, percentile
    '''
    threshold_ub = 0
    for pr in sorted(unique(posterior_pr), reverse=True):
        mbr_mat = posterior_pr >= pr
        if count_nonzero(mbr_mat.any(axis=1)) < len(profiles):
            pass
        else:
            if count_nonzero(mbr_mat.any(axis=0)) == n_groups:
                threshold_ub = pr
                break

    if threshold_ub != 0.0:
        threshold = amin([percentile(posterior_pr, 75), threshold_ub])
        membership_total = posterior_pr >= threshold
    else:
        membership_total = posterior_pr > 0.0
    '''
    membership_total = posterior_pr >= threshold

    from collections import defaultdict
    groups = defaultdict(set)
    for i, membership in enumerate(membership_total):
        for j in nonzero(membership)[0]:
            groups[j].add(profiles.index[i])

    return [frozenset(g) for g in groups.values()]


def gmm(profiles, n_groups, threshold, init='random', n_init=100,
    search_only=False): 
    """Apply the classic Gaussian Mixture Model (GMM) to discover 
    resource groups [1]_.

    This method allows a range of expected number of organizational
    groups to be specified rather than an exact number. It may also act 
    as a helper function for determining a proper selection of number of 
    groups.

    Parameters
    ----------
    profiles : DataFrame
        Constructed resource profiles.
    n_groups : int, or list of ints
        Expected number of resource groups, or a list of candidate
        numbers to be determined.
    threshold : float
        A given threshold value in range [0, 1.0] for producing
        determined clustering from the fuzzy clustering results from GMM.
    init : {'random', 'mja', 'ahc', 'kmeans', 'plain'}, optional, \
        default 'random'
        Options for deciding the strategy for initialization. Could be 
        one of the following:

            - ``'random'``, use random initialization.
            - ``'mja'``, use the mining method Metric based on Joint 
              Activities as initialization method.
            - ``'ahc'``, use the hierarchical method of Agglomerative
              Hierarchical Clustering as initialization method.
            - ``'kmeans'``, use the classic clustering algorithm kMeans 
              as initialization method.
            - ``'plain'``, simply put resources into groups by in the 
              order of their ids.

        Note that if an option other than ``'random'`` is specified, then
        the initialization is performed once only.
    n_init : int, optional, default 100
        Number of times of random initialization (if specified) performs
        before training the clustering model.
    search_only : bool, optional, default False
        A boolean flag indicating whether to search for the number of
        groups only or to perform group discovery based on the search
        result. Defaults to ``False``, i.e., to perform group discovery
        after search searching.

    Returns
    -------
    best_k : int
        The suggested selection of number of groups (if `search_only` is
        True).
    list of frozensets
        Discovered resource groups (if `search_only` is False).

    Raises
    ------
    TypeError
        If the parameter type for `n_groups` is unexpected.

    See Also
    --------
    sklearn.mixture.GaussianMixture
    orgminer.OrganizationalModelMiner.community.graph_partitioning.mja
    orgminer.OrganizationalModelMiner.clustering.hierarchical.ahc
    sklearn.cluster.KMeans

    References
    ----------
    .. [1] Yang, J., Ouyang, C., Pan, M., Yu, Y., & ter Hofstede, A.
       (2018). Finding the Liberos: Discover Organizational Models with
       Overlaps. In *Proceedings of the 16th International Conference on
       Business Process Management*, pp. 339-355. Springer, Cham.
       `<https://doi.org/10.1007/978-3-319-98648-7_20>`_
    """
    if type(n_groups) is int:
        return _gmm(profiles, n_groups, threshold, init, n_init)
    elif type(n_groups) is list and len(n_groups) == 1:
        return _gmm(profiles, n_groups[0], threshold, init, n_init)
    elif type(n_groups) is list and len(n_groups) > 1:
        best_k = -1
        best_score = float('-inf')
        from orgminer.OrganizationalModelMiner.utilities import \
            cross_validation_score
        from orgminer.Evaluation.m2m.cluster_validation import silhouette_score
        from numpy import mean, amax
        for k in n_groups:
            # NOTE: use Cross Validation
            score = cross_validation_score(
                X=profiles, miner=_gmm,
                miner_params={
                    'n_groups': k,
                    'threshold': threshold,
                    'init': init,
                    'n_init': n_init
                },
                proximity_metric='euclidean'
            )

            if score > best_score:
                best_score = score
                best_k = k

        print('-' * 80)
        print('Selected "K" = {}'.format(best_k))
        if search_only:
            return best_k
        else:
            return _gmm(profiles, best_k, threshold, init, n_init)
    else:
        raise TypeError('Invalid type for parameter ``{}``: {}'.format(
            'n_groups', type(n_groups)))


def _moc(profiles, n_groups, init='random', n_init=100):
    """Apply the Model-based Overlapping Clustering (MOC) to discover 
    resource groups [1]_.

    Parameters
    ----------
    profiles : DataFrame
        Constructed resource profiles.
    n_groups : int
        Expected number of resource groups.
    init : {'random', 'mja', 'ahc', 'kmeans', 'plain'}, optional, \
        default 'random'
        Options for deciding the strategy for initialization. Could be 
        one of the following:

            - ``'random'``, use random initialization.
            - ``'mja'``, use the mining method Metric based on Joint 
              Activities as initialization method.
            - ``'ahc'``, use the hierarchical method of Agglomerative
              Hierarchical Clustering as initialization method.
            - ``'kmeans'``, use the classic clustering algorithm kMeans 
              as initialization method.
            - ``'plain'``, simply put resources into groups by in the 
              order of their ids.

        Note that if an option other than ``'random'`` is specified, then
        the initialization is performed once only.
    n_init : int, optional, default 100
        Number of times of random initialization (if specified) performs
        before training the clustering model.

    Returns
    -------
    list of frozensets
        Discovered resource groups.

    Raises
    ------
    ValueError
        If the specified option for initialization is invalid.
    RuntimeError
        If no valid result could be produced.

    See Also
    --------
    orgminer.OrganizationalModelMiner.community.graph_partitioning.mja
    orgminer.OrganizationalModelMiner.clustering.hierarchical.ahc
    sklearn.cluster.KMeans

    References
    ----------
    .. [1] Yang, J., Ouyang, C., Pan, M., Yu, Y., & ter Hofstede, A.
       (2018). Finding the Liberos: Discover Organizational Models with
       Overlaps. In *Proceedings of the 16th International Conference on
       Business Process Management*, pp. 339-355. Springer, Cham.
       `<https://doi.org/10.1007/978-3-319-98648-7_20>`_
    """
    print('Applying overlapping clustering-based MOC:')
    # step 0. Perform specific initialization method (if given)
    if init in ['mja', 'ahc', 'kmeans', 'plain']:
        warm_start = True
        from numpy import zeros
        from pandas import DataFrame
        if init == 'mja':
            from orgminer.OrganizationalModelMiner.community.graph_partitioning\
                import _mja
            init_groups = _mja(profiles, n_groups)  
        elif init == 'ahc':
            from .hierarchical import _ahc
            init_groups, _ = _ahc(profiles, n_groups)
        elif init == 'kmeans':
            init_groups = list(set() for i in range(n_groups))
            from sklearn.cluster import KMeans
            labels = KMeans(n_clusters=n_groups, random_state=0).fit_predict(
                profiles)
            for i, r in enumerate(sorted(profiles.index)):
                init_groups[labels[i]].add(r)
        else: # init == 'plain'
            init_groups = list(set() for i in range(n_groups))
            for i, r in enumerate(sorted(profiles.index)):
                init_groups[i % n_groups].add(r)
        print('Initialization done using {}:'.format(init))

        m = DataFrame(zeros((len(profiles), n_groups)), index=profiles.index)
        for i, g in enumerate(init_groups):
            for r in g:
                m.loc[r][i] = 1 # set the membership matrix as init input
    elif init == 'random':
        warm_start = False
    else:
        raise ValueError('Invalid value for parameter `{}`: {}'.format(
            'init', init))

    # step 1. Train the model
    from .classes import MOC
    if warm_start:
        moc_model = MOC(n_components=n_groups, M_init=m.values, n_init=1)
    else:
        moc_model = MOC(n_components=n_groups, n_init=n_init)

    # step 2. Derive the clusters as the end result
    mat_membership = moc_model.fit_predict(profiles.values)

    from numpy import nonzero
    from collections import defaultdict
    groups = defaultdict(set)
    for i in range(len(mat_membership)):
        # check if any valid membership exists for the resource based on
        # the results predicted by the obtained MOC model
        if mat_membership[i,:].any(): # valid if at least belongs to 1 group
            for j in nonzero(mat_membership[i,:])[0]:
                groups[j].add(profiles.index[i])
        else: # invalid (unexpected exit)
            print(mat_membership)
            raise RuntimeError('No valid result could be produced.')

    return [frozenset(g) for g in groups.values()]


def moc(profiles, n_groups, init='random', n_init=100,
    search_only=False):
    """Apply the Model-based Overlapping Clustering (MOC) to discover 
    resource groups [1]_.

    This method allows a range of expected number of organizational
    groups to be specified rather than an exact number. It may also act 
    as a helper function for determining a proper selection of number of 
    groups.

    Parameters
    ----------
    profiles : DataFrame
        Constructed resource profiles.
    n_groups : int, or list of ints
        Expected number of resource groups, or a list of candidate
        numbers to be determined.
    init : {'random', 'mja', 'ahc', 'kmeans', 'plain'}, optional, \
        default 'random'
        Options for deciding the strategy for initialization. Could be 
        one of the following:

            - ``'random'``, use random initialization.
            - ``'mja'``, use the mining method Metric based on Joint 
              Activities as initialization method.
            - ``'ahc'``, use the hierarchical method of Agglomerative
              Hierarchical Clustering as initialization method.
            - ``'kmeans'``, use the classic clustering algorithm kMeans 
              as initialization method.
            - ``'plain'``, simply put resources into groups by in the 
              order of their ids.

        Note that if an option other than ``'random'`` is specified, then
        the initialization is performed once only.
    n_init : int, optional, default 100
        Number of times of random initialization (if specified) performs
        before training the clustering model.
    search_only : bool, optional, default False
        A boolean flag indicating whether to search for the number of
        groups only or to perform group discovery based on the search
        result. Defaults to ``False``, i.e., to perform group discovery
        after searching.

    Returns
    -------
    best_k : int
        The suggested selection of number of groups (if `search_only` is
        True).
    list of frozensets
        Discovered resource groups (if `search_only` is False).

    Raises
    ------
    TypeError
        If the parameter type for `n_groups` is unexpected.

    See Also
    --------
    orgminer.OrganizationalModelMiner.community.graph_partitioning.mja
    orgminer.OrganizationalModelMiner.clustering.hierarchical.ahc
    sklearn.cluster.KMeans

    References
    ----------
    .. [1] Yang, J., Ouyang, C., Pan, M., Yu, Y., & ter Hofstede, A.
       (2018). Finding the Liberos: Discover Organizational Models with
       Overlaps. In *Proceedings of the 16th International Conference on
       Business Process Management*, pp. 339-355. Springer, Cham.
       `<https://doi.org/10.1007/978-3-319-98648-7_20>`_
    """
    if type(n_groups) is int:
        return _moc(profiles, n_groups, init, n_init)
    elif type(n_groups) is list and len(n_groups) == 1:
        return _moc(profiles, n_groups[0], init, n_init)
    elif type(n_groups) is list and len(n_groups) > 1:
        best_k = -1
        best_score = float('-inf')
        from orgminer.OrganizationalModelMiner.utilities import \
            cross_validation_score
        from orgminer.Evaluation.m2m.cluster_validation import silhouette_score
        from numpy import mean, amax
        for k in n_groups:
            # NOTE: use Cross Validation
            score = cross_validation_score(
                X=profiles, miner=_moc,
                miner_params={
                    'n_groups': k,
                    'init': init,
                    'n_init': n_init
                },
                proximity_metric='euclidean'
            )

            if score > best_score:
                best_score = score
                best_k = k

        print('-' * 80)
        print('Selected "K" = {}'.format(best_k))
        if search_only:
            return best_k
        else:
            return _moc(profiles, best_k, init, n_init)
    else:
        raise TypeError('Invalid type for parameter `{}`: {}'.format(
            'n_groups', type(n_groups)))


def _fcm(profiles, n_groups, threshold, init='random', n_init=100): 
    """Apply the Fuzzy C-Means (FCM) [1]_ to discover resource groups.

    Parameters
    ----------
    profiles : DataFrame
        Constructed resource profiles.
    n_groups : int
        Expected number of resource groups.
    threshold : float
        A given threshold value in range [0, 1.0] for producing
        determined clustering from the fuzzy clustering results from FCM.
    init : {'random', 'mja', 'ahc', 'kmeans', 'plain'}, optional, \
        default 'random'
        Options for deciding the strategy for initialization. Could be 
        one of the following:

            - ``'random'``, use random initialization.
            - ``'mja'``, use the mining method Metric based on Joint 
              Activities as initialization method.
            - ``'ahc'``, use the hierarchical method of Agglomerative
              Hierarchical Clustering as initialization method.
            - ``'kmeans'``, use the classic clustering algorithm kMeans 
              as initialization method.
            - ``'plain'``, simply put resources into groups by in the 
              order of their ids.

        Note that if an option other than ``'random'`` is specified, then
        the initialization is performed once only.
    n_init : int, optional, default 100
        Number of times of random initialization (if specified) performs
        before training the clustering model.

    Returns
    -------
    list of frozensets
        Discovered resource groups.

    Raises
    ------
    ValueError
        If the specified option for initialization is invalid.

    See Also
    --------
    orgminer.OrganizationalModelMiner.community.graph_partitioning.mja
    orgminer.OrganizationalModelMiner.clustering.hierarchical.ahc
    sklearn.cluster.KMeans

    References
    ----------
    .. [1] Tan, P. N., Steinbach, M., Karpatne, A., & Kumar, V. (2018).
       *Introduction to Data Mining*.
    """
    print('Applying overlapping clustering-based FCM:')
    # step 0. Perform specific initialization method (if given)
    if init in ['mja', 'ahc', 'kmeans', 'plain']:
        warm_start = True
        if init == 'mja':
            from orgminer.OrganizationalModelMiner.community.graph_partitioning\
                import _mja
            init_groups = _mja(profiles, n_groups)  
        elif init == 'ahc':
            from .hierarchical import _ahc
            init_groups, _ = _ahc(profiles, n_groups)
        elif init == 'kmeans':
            init_groups = list(set() for i in range(n_groups))
            from sklearn.cluster import KMeans
            labels = KMeans(n_clusters=n_groups, random_state=0).fit_predict(
                profiles)
            for i, r in enumerate(sorted(profiles.index)):
                init_groups[labels[i]].add(r)
        else: # init == 'plain'
            init_groups = list(set() for i in range(n_groups))
            for i, r in enumerate(sorted(profiles.index)):
                init_groups[i % n_groups].add(r)
        print('Initialization done using {}:'.format(init))
    elif init == 'random':
        warm_start = False
    else:
        raise ValueError('Invalid value for parameter `{}`: {}'.format(
            'init', init))

    # step 1. Train the model
    from .classes import FCM
    if warm_start:
        from numpy import array, mean, nonzero, zeros
        init_means = list()
        for g in init_groups:
            init_means.append(mean(profiles.loc[list(g)].values, axis=0))
        fcm_model = FCM(
            n_components=n_groups,
            n_init=1,
            means_init=array(init_means))
    else:
        fcm_model = FCM(
            n_components=n_groups,
            n_init=n_init)

    # step 2. Derive the clusters as the end result
    fpp = fcm_model.fit_predict(profiles.values)

    from numpy import array, nonzero, argmax, median
    from numpy.random import choice
    from collections import defaultdict
    groups = defaultdict(set)
    for i, belonging_factor in enumerate(fpp):
        if threshold is None:
            #threshold = median(belonging_factor[belonging_factor != 0])
            threshold = 1.0 / n_groups
        membership = array([p >= threshold for p in belonging_factor])

        # check if any valid membership exists for the resource based on
        # the selection of the threshold
        if membership.any():
            for j in nonzero(membership)[0]:
                groups[j].add(profiles.index[i])
        else: # invalid, have to choose the maximum one or missing the resource
            groups[argmax(belonging_factor)].add(profiles.index[i])

    return [frozenset(g) for g in groups.values()]


def fcm(profiles, n_groups, threshold, init='random', n_init=100,
    search_only=False): 
    """Apply the Fuzzy C-Means (FCM) [1]_ to discover resource groups.

    This method allows a range of expected number of organizational
    groups to be specified rather than an exact number. It may also act 
    as a helper function for determining a proper selection of number of 
    groups.

    Parameters
    ----------
    profiles : DataFrame
        Constructed resource profiles.
    n_groups : int, or list of ints
        Expected number of resource groups, or a list of candidate
        numbers to be determined.
    threshold : float
        A given threshold value in range [0, 1.0] for producing
        determined clustering from the fuzzy clustering results from FCM.
    init : {'random', 'mja', 'ahc', 'kmeans', 'plain'}, optional, \
        default 'random'
        Options for deciding the strategy for initialization. Could be 
        one of the following:

            - ``'random'``, use random initialization.
            - ``'mja'``, use the mining method Metric based on Joint 
              Activities as initialization method.
            - ``'ahc'``, use the hierarchical method of Agglomerative
              Hierarchical Clustering as initialization method.
            - ``'kmeans'``, use the classic clustering algorithm kMeans 
              as initialization method.
            - ``'plain'``, simply put resources into groups by in the 
              order of their ids.

        Note that if an option other than ``'random'`` is specified, then
        the initialization is performed once only.
    n_init : int, optional, default 100
        Number of times of random initialization (if specified) performs
        before training the clustering model.
    search_only : bool, optional, default False
        A boolean flag indicating whether to search for the number of
        groups only or to perform group discovery based on the search
        result. Defaults to ``False``, i.e., to perform group discovery
        after searching.

    Returns
    -------
    best_k : int
        The suggested selection of number of groups (if `search_only` is
        True).
    list of frozensets
        Discovered resource groups (if `search_only` is False).

    Raises
    ------
    TypeError
        If the parameter type for `n_groups` is unexpected.

    See Also
    --------
    orgminer.OrganizationalModelMiner.community.graph_partitioning.mja
    orgminer.OrganizationalModelMiner.clustering.hierarchical.ahc
    sklearn.cluster.KMeans

    References
    ----------
    .. [1] Tan, P. N., Steinbach, M., Karpatne, A., & Kumar, V. (2018).
       *Introduction to Data Mining*.
    """
    if type(n_groups) is int:
        return _fcm(profiles, n_groups, threshold, init, n_init)
    elif type(n_groups) is list and len(n_groups) == 1:
        return _fcm(profiles, n_groups[0], threshold, init, n_init)
    elif type(n_groups) is list and len(n_groups) > 1:
        from orgminer.OrganizationalModelMiner.utilities import \
            cross_validation_score
        best_k = -1
        best_score = float('-inf')
        for k in n_groups:
            score = cross_validation_score(
                X=profiles, miner=_fcm,
                miner_params={
                    'n_groups': k,
                    'threshold': threshold,
                    'init': init,
                    'n_init': n_init
                },
                proximity_metric='euclidean'
            )
            if score > best_score:
                best_score = score
                best_k = k

        print('-' * 80)
        print('Selected "K" = {}'.format(best_k))
        if search_only:
            return best_k
        else:
            return _fcm(profiles, best_k, threshold, init, n_init)
    else:
        raise TypeError('Invalid type for parameter `{}`: {}'.format(
            'n_groups', type(n_groups)))




# -*- coding: utf-8 -*-

"""This module provides the necessary utilities for processing data 
structures related to social network mining.
"""
def select_edges_by_weight(sn, weight='weight',
    low=None, high=None, 
    percentage=None):
    '''Select and keep edges in an input social network based on edge 
    weight values. 
    
    Notes
    -----
    The filtering process can be done in either of the two ways:
        
        1. Define an interval [`low`, `high`) and keep only edges 
        with weight values falling within the interval. If either one of 
        the two sides of bound is not defined, infinity is used instead.
        
        2. Specify a percentage value and keep only edges with weight
        values within the range, e.g., with ``'+0.75'`` specified, only 
        edges with the highest 75% weight values are kept.

    Note that one of the two ways must be specified, and the first would 
    override the second if both are presented.

    Parameters
    ----------
    sn : NetworkX (Di)Graph
        A social network.
    weight : str, optional, default 'weight'
        Name of the weight attribute in the network. Defaults to 
        ``'weight'``, which is the default weight attribute name used by 
        NetworkX.
    low : float, optional
        The lower bound value, used for the first way of filtering.
    high : float, optional
        The upper bound value, used for the first way of filtering.
    percentage : str, optional
        The specified percentage value, used for the second way of 
        filtering. Should be formatted as a plus/minus sign concatenated 
        with a float number in range [0, 1), e.g., ``+0.75``.

    Returns
    -------
    psn: NetworkX (Di)Graph
        The social network with certain edges filtered out.
    '''

    from copy import deepcopy
    psn = deepcopy(sn)
    n_edges_old = len(psn.edges)
    edges_to_be_removed = list()
    if low is not None or high is not None:
        test_lower_bound = lambda wt: wt >= low if low is not None else True
        test_upper_bound = lambda wt: wt < high if high is not None else True
        print('\tOnly values within [{}, {}) are kept.'.format(low, high))
        for u, v, wt in psn.edges.data(weight):
            if not (test_lower_bound(wt) and test_upper_bound(wt)):
                edges_to_be_removed.append((u, v))
    elif percentage is not None:
        from numpy import percentile
        if percentage[0] == '+':
            # keep the top p ranked edges
            q = 1 - float(percentage[1:])
            print('\tOnly top {}% are kept.'.format((1 - q) * 100))
            threshold = percentile([e[2] 
                for e in psn.edges.data(weight)], q * 100)
            test_condition = lambda wt: wt >= threshold
        elif percentage[0] == '-':
            # keep the bottom p ranked edges
            q = float(percentage[1:])
            print('\tOnly bottom {}% are kept.'.format(q * 100))
            threshold = percentile([e[2] 
                for e in psn.edges.data(weight)], q * 100)
            test_condition = lambda wt: wt <= threshold
        else:
            exit('Error processing due to invalid input.')
        
        for u, v, wt in psn.edges.data(weight):
            if not test_condition(wt):
                edges_to_be_removed.append((u, v))
    else:
        raise RuntimeError('No filtering criterion specified.')

    psn.remove_edges_from(edges_to_be_removed)
    print('\t{}/{} edges ({:.2%}) have been removed.'.format(
        len(edges_to_be_removed), n_edges_old,
        len(edges_to_be_removed) / n_edges_old))

    return deepcopy(psn)




# -*- coding: utf-8 -*-

"""This module contains the implementation of methods for profiling 
resources using information directly accessible in a resource log.
"""
def count_execution_frequency(rl, scale=None):
    """Build resource profiles based on how frequently resources 
    originated events of execution modes.
    
    Each column in the result profiles corresponds with an execution
    mode captured in the given resource log.

    Parameters
    ----------
    rl : DataFrame
        A resource log.
    scale : {None, 'normalize', log'}, optional, default None
        Options for deciding how to scale the values of frequency
        counting. Could be one of the following:
            
            - ``None``, no scaling will be performed.
            - ``'normalize'``, scale the frequency values by the total
              count of executions by each resource (scale by row).
            - ``'log'``, scale the frequency values by logarithm.

    Returns
    -------
    DataFrame
        The constructed resource profiles.
    
    Raises
    ------
    ValueError
        If the specified option for scaling is invalid.
    """
    from collections import defaultdict
    mat = defaultdict(lambda: defaultdict(lambda: 0))
    for res, trace in rl.groupby('resource'):
        for event in trace.itertuples():
            exec_mode = (event.case_type, event.activity_type, event.time_type)
            mat[res][exec_mode] += 1

    from pandas import DataFrame
    df = DataFrame.from_dict(mat, orient='index').fillna(0)
    if scale is None:
        return df
    elif scale == 'normalize':
        return df.div(df.sum(axis=1), axis=0)
    elif scale == 'log':
        # NOTE: log_e(x + 1)
        from numpy import log 
        return df.apply(lambda x: log(x + 1))
    else:
        raise ValueError('Invalid value for parameter `{}`: {}'.format(
            'scale', scale))




# -*- coding: utf-8 -*-

"""This module contains the implementation of hierarchical organizational 
mining methods, based on the use of clustering techniques.
"""
def _ahc(profiles, n_groups, method='single', metric='euclidean'):
    """Apply the classic Agglomerative Hierarchical Clustering (AHC) 
    [1]_ to discover resource groups.

    Parameters
    ----------
    profiles : DataFrame
        Constructed resource profiles.
    n_groups : int
        Expected number of resource groups.
    method : str, optional, default 'single'
        Choice of methods for merging clusters at each iteration.
    metric : str, optional, default 'euclidean'
        Choice of metrics for measuring the distance while calculating 
        distance. Defaults to ``'euclidean'``, meaning that euclidean
        distance is used for measuring distance.

    Returns
    -------
    list of frozensets
        Discovered resource groups.
    og_hcy : DataFrame
        The hierarchical structure (dendrogram) as a pandas DataFrame,
        with resource ids as indices, levels of the hierarchy as columns,
        and group ids as the values. 
        E.g. for 20 resources placed in a 5-level hierarchical structure
        with 8 groups at the lowest level, there should be 20 rows and 5
        columns in the DataFrame, and the values should be in range of 0
        to 7.
    
    See Also
    --------
    scipy.cluster.hierarchy
    scipy.spatial.distance
    pandas.DataFrame

    References
    ----------
    .. [1] Song, M., & van der Aalst, W. M. P. (2008). Towards
       comprehensive support for organizational mining. *Decision Support
       Systems*, 46(1), 300-317.
       `<https://doi.org/10.1016/j.dss.2008.07.002>`_
    """
    print('Applying hierarchical clustering-based AHC:')
    from scipy.cluster import hierarchy
    Z = hierarchy.linkage(profiles, method=method, metric=metric)
    # the hierachical tree as a matrix where each column corresponds to a
    # specific level
    mx_tree = hierarchy.cut_tree(Z, n_clusters=range(1, n_groups + 1))
    # wrap as DataFrame og_hcy
    from pandas import DataFrame
    og_hcy = DataFrame(mx_tree, index=profiles.index)

    from collections import defaultdict
    groups = defaultdict(set)
    # add by each resource
    for i in range(len(og_hcy.index)):
        groups[og_hcy.iloc[i,-1]].add(og_hcy.index[i])

    return [frozenset(g) for g in groups.values()], og_hcy


def ahc(profiles, n_groups, method='single', metric='euclidean',
    search_only=False):
    """Apply the classic Agglomerative Hierarchical Clustering (AHC) 
    [1]_ to discover resource groups.
    
    This method allows a range of expected number of organizational
    groups to be specified rather than an exact number. It may also act 
    as a helper function for determining a proper selection of number of 
    groups.

    Parameters
    ----------
    profiles : DataFrame
        Constructed resource profiles.
    n_groups : int, or list of ints
        Expected number of resource groups, or a list of candidate
        numbers to be determined.
    method : str, optional, default 'single'
        Choice of methods for merging clusters at each iteration.
    metric : str, optional, default 'euclidean'
        Choice of metrics for measuring the distance while calculating 
        distance. Defaults to ``'euclidean'``, meaning that euclidean
        distance is used for measuring distance.
    search_only : bool, optional, default False
        A boolean flag indicating whether to search for the number of
        groups only or to perform cluster analysis based on the search
        result. Defaults to ``False``, i.e., to perform cluster analysis
        after searching.

    Returns
    -------
    best_k : int
        The suggested selection of number of groups (if `search_only` 
        is True).
    best_ogs : list of frozensets
        Discovered resource groups (if `search_only` is False).
    best_og_hcy : DataFrame
        The hierarchical structure (dendrogram) as a pandas DataFrame,
        with resource ids as indices, levels of the hierarchy as columns,
        and group ids as the values (if `search_only` is False).
        E.g. for 20 resources placed in a 5-level hierarchical structure
        with 8 groups at the lowest level, there should be 20 rows and 5
        columns in the DataFrame, and the values should be in range of 0
        to 7.

    Raises
    ------
    TypeError
        If the parameter type for `n_groups` is unexpected.

    See Also
    --------
    scipy.cluster.hierarchy
    scipy.spatial.distance
    pandas.DataFrame

    References
    ----------
    .. [1] Song, M., & van der Aalst, W. M. P. (2008). Towards
       comprehensive support for organizational mining. *Decision Support
       Systems*, 46(1), 300-317.
       `<https://doi.org/10.1016/j.dss.2008.07.002>`_
    """
    if type(n_groups) is int:
        return _ahc(profiles, n_groups, method, metric)
    elif type(n_groups) is list and len(n_groups) == 1:
        return _ahc(profiles, n_groups[0], method, metric)
    elif type(n_groups) is list and len(n_groups) > 1:
        best_k = -1
        best_score = float('-inf')
        from orgminer.OrganizationalModelMiner.utilities import \
            cross_validation_score
        from orgminer.Evaluation.m2m.cluster_validation import silhouette_score
        from numpy import mean, amax
        for k in n_groups:
            # NOTE: use Cross Validation
            score = cross_validation_score(
                X=profiles, miner=_ahc,
                miner_params={
                    'n_groups': k,
                    'method': method,
                    'metric': metric 
                },
                proximity_metric = metric
            )

            if score > best_score:
                best_score = score
                best_k = k

        print('-' * 80)
        print('Selected "K" = {}'.format(best_k))
        if search_only:
            return best_k
        else:
            return _ahc(profiles, best_k, method, metric)
    else:
        raise TypeError('Invalid type for parameter `{}`: {}'.format(
            'n_groups', type(n_groups)))




# -*- coding: utf-8 -*-

"""This module contains the implementation of the local diagnostics 
measures.
"""
# NOTE: # (*, rg, q) / (*, rg, *)
def group_relative_focus(group, mode, rl):
    """Measure the relative focus of a group with respect to an execution
    mode.

    Parameters
    ----------
    group : iterator
        Id of resources as a resource group.
    mode : 3-tuple
        An execution mode.
    rl : DataFrame
        A resource log.

    Returns
    -------
    float
        The measured relative focus.
    """
    # filtering irrelevant events
    rl = rl.loc[rl['resource'].isin(group)]
    grouped_by_modes = rl.groupby([
        'case_type', 'activity_type', 'time_type'])
    
    if mode in grouped_by_modes.groups:
        return len(grouped_by_modes.get_group(mode)) / len(rl)
    else:
        return 0.0


# NOTE: # (*, rg, q) / (*, *, q)
def group_relative_stake(group, mode, rl):
    """Measure the relative focus of a group with respect to an execution
    mode.

    Parameters
    ----------
    group : iterator
        Id of resources as a resource group.
    mode : 3-tuple
        An execution mode.
    rl : DataFrame
        A resource log.

    Returns
    -------
    float
        The measured relative focus.
    """
    total_count = len(rl.groupby([
        'case_type', 'activity_type', 'time_type']).get_group(mode))

    # filtering irrelevant events
    rl = rl.loc[rl['resource'].isin(group)]
    grouped_by_modes = rl.groupby([
        'case_type', 'activity_type', 'time_type'])
    if mode in grouped_by_modes.groups:
        return len(grouped_by_modes.get_group(mode)) / total_count
    else:
        return 0.0


# NOTE: R (r, rg, q) / (r, rg)
def group_coverage(group, mode, rl):
    """Measure the coverage of a group with respect to an execution mode.

    Parameters
    ----------
    group : iterator
        Id of resources as a resource group.
    mode : 3-tuple
        An execution mode.
    rl : DataFrame
        A resource log.

    Returns
    -------
    float
        The measured coverage.
    """
    # filtering irrelevant events
    rl = rl.loc[rl['resource'].isin(group)]

    num_participants = 0
    for r in group:
        if len(rl.loc[
            (rl['resource'] == r) &
            (rl['case_type'] == mode[0]) &
            (rl['activity_type'] == mode[1]) &
            (rl['time_type'] == mode[2])]) > 0:
            num_participants += 1
        else:
            pass
    return num_participants / len(group)


# NOTE: # (r, rg, q) / (*, rg, q)
def group_member_contribution(group, mode, rl):
    # filtering irrelevant events
    rl = rl.loc[rl['resource'].isin(group)].groupby([
        'case_type', 'activity_type', 'time_type']).get_group(mode)
    group_total_count = len(rl)

    group_load_distribution = dict()
    for r in group:
        group_load_distribution[r] = (
            len(rl.loc[rl['resource'] == r]) / group_total_count)

    return group_load_distribution




# -*- coding: utf-8 -*-

"""This module contains methods for importing event log data from files 
in specific formats.

Event log formats currently supported:

    - Disco-exported CSV format (https://fluxicon.com/disco/)
    - eXtensible Event Stream (XES) (http://xes-standard.org/)

All methods in this module should return the successfully imported event 
log as a pandas DataFrame.

Notes
-----
The original event log data should at least provide the information of 
the activity ids. It is expected that the resource ids and timestamps 
are presented as well (not mandatory though).
Other event data attributes can be appended according to different 
purposes of working projects and event data available.

See Also
--------
pandas.DataFrame : The primary pandas data structure.
"""
def _describe_event_log(el):
    """Output descriptive information for a successfully imported event
    log.

    Parameters
    ----------
    el : DataFrame
        An event log imported.

    Returns
    -------
    """
    print('-' * 80)
    print('Number of events:\t\t{}'.format(len(el)))
    print('Number of cases:\t\t{}'.format(len(el.groupby('case_id'))))
    #print('Event log attributes:\n\t')
    print('-' * 80)


def read_disco_csv(f, header=True):
    """Import an event log from a file in CSV (Column-Separated Values)
    format, exported from Disco.

    There are four expected "default" event attributes, including:

        - case_id,
        - activity,
        - resource,
        - timestamp,

    that start as the first four columns.

    Parameters
    ----------
    f : File object
        File object of the event log being imported.
    header : bool, optional, default True
        A boolean flag indicating whether the input event log file
        contains a header line. Defaults to ``True``, i.e., the provided 
        CSV file is expected to be having a header line.

    Returns
    -------
    el : DataFrame
        An event log.
    """
    from csv import reader

    is_header_line = header

    ld = list()
    line_count = 0

    # default attributes consistent with Disco export function
    attributes = {
        'case_id': 0,
        'activity': 1,
        'resource': 2,
        'timestamp': 3
    }
    num_default_attributes = len(attributes)
    has_additional_attributes = False
    attributes_registered = False

    for row in reader(f):
        if not row:
            continue

        if not attributes_registered:
            has_additional_attributes = len(row) > num_default_attributes
            if has_additional_attributes:
                # register them
                if is_header_line:
                    for i, name in enumerate(row[num_default_attributes:]):
                        col_index = i + num_default_attributes
                        attributes[name] = col_index
                else:
                    # no header line scanned but additional attributes exist
                    for i in range(len(row[num_default_attributes:])):
                        col_index = i + num_default_attributes
                        name = 'unknown_attr_col{}'.format(col_index + 1)
                        attributes[name] = col_index
            else:
                pass
            attributes_registered = True

        if is_header_line:
            is_header_line = False
        else:
            e = dict()
            for attr, col_index in attributes.items():
                e[attr] = row[col_index]
            ld.append(e)

        line_count += 1

    from pandas import DataFrame
    el = DataFrame(ld)

    print('Imported successfully. {} lines scanned.'.format(line_count))

    _describe_event_log(el)
    return el


def read_xes(f):
    """Import an event log from a file in IEEE XES (eXtensible Event 
    Stream) format.

    This is a wrapper method around the XES import feature provided in
    the PM4Py library. See: `<http://pm4py.org/>`_

    There are four expected default event attributes, including:

        - case_id (mapped from `case:concept:name`),
        - activity (mapped from `concept:name`),
        - resource (mapped from `org:resource`),
        - timestamp (mapped from `time:timestamp`).

    Parameters
    ----------
    f : File object
        File object of the event log being imported.

    Returns
    -------
    el : DataFrame
        An event log.
    """
    from pm4py.objects.log.importer.xes import factory
    pm4py_log = factory.import_log_from_string(f.read(), 
        parameters={'index_trace_indexes': True})
    from pm4py.objects.conversion.log.versions import to_dataframe
    df = to_dataframe.apply(pm4py_log).rename(columns={
        'case:concept:name': 'case_id',
        'concept:name': 'activity',
        'org:resource': 'resource',
        'time:timestamp': 'timestamp'
    })
    import re
    df['timestamp'] = df['timestamp'].apply(
        lambda x: x.strftime('%Y/%m/%d %H:%M:%S.%f'))
    
    _describe_event_log(df)
    return df




# -*- coding: utf-8 -*-

"""This module contains the implementation of methods for mining social 
networks from an event log, using metrics based on causality [1]_.

See Also
--------
orgminer.SocialNetworkMiner.joint_activities
orgminer.SocialNetworkMiner.joint_cases

Notes
-----
Including real causality into consideration requires the presence of a
process model related to the given event log.

References
----------
.. [1] Van der Aalst, W. M. P., Reijers, H. A., & Song, M. (2005).
   Discovering social networks from event logs. *Computer Supported
   Cooperative Work (CSCW)*, 14(6), 549-593.
   `<https://doi.org/10.1007/s10606-005-9005-9>`_
"""
def handover(el,
    real_causality, direct_succession, multiple_transfers,
    depth=1, beta=1):
    """Discover a social network from an event log based on handover of 
    work metric, which considers the relationship between a pair of 
    resources completing subsequent activities in process execution.

    Parameters
    ----------
    el : DataFrame
        An event log.
    real_causality : bool
        A boolean flag indicating whether to consider arbitrary 
        transfers of work, or only those for which there is a causal 
        dependency according to the process model.
    direct_succession : bool
        A boolean flag indicating whether to consider the degree of 
        causality, i.e., direct or indirect succession.
    multiple_transfers : bool
        A boolean flag indicating whether to consider multiple transfers 
        within one case or not.
    depth : int, optional, default 1
        The degree of causality to be considered. Only use when 
        `direct_succession` is False, i.e., indirect succession 
        considered.
    beta : float, optional, default 1
        The causality fall factor.

    Returns
    -------
    sn : NetworkX DiGraph
        The discovered social network.
    """
    from collections import defaultdict
    mat = defaultdict(lambda: defaultdict(lambda: {'weight': 0.0}))
    if direct_succession and multiple_transfers: # CDCM
        # scale_factor: SIGMA_Case c in Log (|c| - 1)
        sf = sum(len(trace) - 1 for case_id, trace in el.groupby('case_id'))
        for case_id, trace in el.groupby('case_id'):
            for i in range(len(trace) - 1):
                res_prev = trace.iloc[i]['resource']
                res_next = trace.iloc[i + 1]['resource']
                if res_prev != res_next: # self-loop ignored
                    mat[res_prev][res_next]['weight'] += 1 / sf

    elif direct_succession and not multiple_transfers: # CDIM
        # scale_factor: |L|
        sf = len(el.groupby('case_id'))
        for case_id, trace in el.groupby('case_id'):
            handovered_pairs = set()
            for i in range(len(trace) - 1):
                res_prev = trace.iloc[i]['resource']
                res_next = trace.iloc[i + 1]['resource']
                if res_prev != res_next: # self-loop ignored
                    handovered_pairs.add((res_prev, res_next))
            for pair in handovered_pairs:
                mat[pair[0]][pair[1]]['weight'] += 1 / sf

    elif not direct_succession and multiple_transfers: # CICM 
        # scale_factor: SIGMA_Case c in 
        # Log (SIGMA_n=1:min(|c| - 1, depth) (beta^n-1 * (|c| - n)))
        sf = 0
        for case_id, trace in el.groupby('case_id'):
            num_events = len(trace)
            sf += (sum(beta ** (n - 1) * (num_events - n) 
                for n in range(1, min(num_events - 1, depth))))
        for case_id, trace in el.groupby('case_id'):
            # for each case
            num_events = len(trace)
            for i in range(num_events - 1):
                for n in range(1, min(num_events - 1, depth)):
                    # for each level of calculation depth
                    if i + n <= num_events - 1:
                        res_prev = trace.iloc[i]['resource']
                        res_next = trace.iloc[i + 1]['resource']
                        if res_prev != res_next: # self-loop ignored
                            mat[res_prev][res_next]['weight'] += (
                                beta ** (n - 1) * 1 / sf)

    else: # CIIM
        # scale_factor: SIGMA_Case c in 
        # Log (SIGMA_n=1:min(|c| - 1, depth) (beta^n-1))
        sf = 0
        for case_id, trace in el.groupby('case_id'):
            num_events = len(trace)
            sf += (sum(beta ** (n - 1) 
                for n in range(1, min(num_events - 1, depth))))
        for case_id, trace in el.groupby('case_id'):
            # for each case
            handovered_pairs = set()
            num_events = len(trace)
            for i in range(num_events - 1):
                for n in range(1, min(num_events - 1, depth)):
                    # for each level of calculation depth
                    if i + n <= num_events - 1:
                        res_prev = trace.iloc[i]['resource']
                        res_next = trace.iloc[i + 1]['resource']
                        if res_prev != res_next: # self-loop ignored
                            handovered_pairs.add((res_prev, res_next, 
                                beta ** (n - 1)))
            for pair in handovered_pairs:
                mat[pair[0]][pair[1]]['weight'] += pair[2] * 1 / sf

    from networkx import DiGraph
    sn = DiGraph(mat)
    sn.add_nodes_from(el.groupby('resource').groups.keys())
    return sn


# TODO: Subcontracting metric
def subcontracting():
    raise NotImplementedError




# -*- coding: utf-8 -*-

"""This module contains methods for determining execution modes 
associated with resource groups.
"""
from orgminer.OrganizationalModelMiner.base import OrganizationalModel

def full_recall(groups, rl):
    """Assign an execution mode to a group, as long as there exists a 
    member resource who has executed the mode.

    Parameters
    ----------
    groups : list of sets
        Resource groups containing resource ids.
    rl : DataFrame
        A resource log.

    Returns
    -------
    om : OrganizationalModel
        An organizational model resulted from assigning execution modes 
        to resource groups.
    """
    print('Applying FullRecall for mode assignment:')
    grouped_by_resource = rl.groupby('resource')
    om = OrganizationalModel()

    for group in groups:
        modes = set()
        for r in group:
            # TODO: optimize the update
            modes.update((e.case_type, e.activity_type, e.time_type)
                for e in grouped_by_resource.get_group(r).itertuples())
        om.add_group(group, sorted(list(modes)))
    return om


#NOTE:: Implementation #1 - OverallScore-WA
def overall_score(groups, rl, p=0.5, w1=0.5, w2=None, auto_search=False):
    """Assign an execution mode to a group, as long as the overall score
    (as a weighted average) of its group relative stake and member
    coverage, is higher than a given threshold.

    Parameters
    ----------
    groups : list of sets
        Resource groups containing resource ids.
    rl : DataFrame
        A resource log.
    p : float, optional, default 0.5
        A given threshold value in range [0, 1.0].
    w1 : float, optional, default 0.5
        The weight value assigned to participation rate.
    w2 : float, optional, default None
        The weight value assigned to coverage. Note that this parameter
        is in fact redundant as its value must conform with the value set
        for parameter `w1`. See notes below.
    auto_search : bool, optional, default False
        A Boolean flag indicating whether to perform auto grid search to
        determine the threshold and the weightings. When auto grid search
        is required, values in range [0, 1.0] will be tested at a step of 
        0.1, and values given to parameter `p`, `w1`, and `w2` will be
        overridden. Defaults to ``False``, i.e., auto grid search is not 
        to be performed.

    Returns
    -------
    om : OrganizationalModel
        An organizational model resulted from assigning execution modes 
        to resource groups.

    Notes
    -----
        The weighting values are expected to sum up to 1.
    """
    if auto_search is True:
        print('Applying OverallScore with auto search:')
        from orgminer.OrganizationalModelMiner.utilities import grid_search
        from functools import partial
        from copy import deepcopy
        from orgminer.Evaluation.l2m.conformance import f1_score
        solution, params = grid_search(
            partial(overall_score, groups=deepcopy(groups), rl=rl), 
            params_config={
            'p': list(x / 10 for x in range(1, 10)),
            'w1': list(x / 10 for x in range(1, 10))},
            func_eval_score=partial(f1_score, rl)
        )
        print('\tBest solution obtained with parameters:')
        print('\t', end='')
        print(params)
        return solution
    else:
        w2 = 1.0 - w1
        print('Applying OverallScore with weights ({}, {}) '.format(w1, w2) +
            'and threshold {} '.format(p) + 'for mode assignment:')

    all_execution_modes = set(rl[['case_type', 'activity_type', 'time_type']]
        .drop_duplicates().itertuples(index=False, name=None))

    from orgminer.Evaluation.l2m.diagnostics import \
        group_relative_stake, group_coverage
    from collections import defaultdict

    scores_group_rel_stake = defaultdict(lambda: defaultdict(dict))
    scores_group_cov = defaultdict(lambda: defaultdict(dict))

    # obtain scores
    for i, group in enumerate(groups):
        for m in all_execution_modes:
            rel_stake = group_relative_stake(group, m, rl)
            scores_group_rel_stake[i][m] = rel_stake
            cov = group_coverage(group, m, rl)
            scores_group_cov[i][m] = cov
    
    # assign based on threshold filtering
    from operator import itemgetter
    om = OrganizationalModel()
    for i, group in enumerate(groups):
        tmp_modes = list()
        for m in all_execution_modes:
            relatedness_score = (
                w1 * scores_group_rel_stake[i][m] +
                w2 * scores_group_cov[i][m]
            )
            if relatedness_score > p:
                tmp_modes.append((m, relatedness_score))
        modes = list(item[0] 
            for item in sorted(tmp_modes, key=itemgetter(1), reverse=True))
        om.add_group(group, modes)

    return om


#NOTE: Implementation #2 - OverallScore-HM
def _overall_score(groups, rl, p):
    """Assign an execution mode to a group, as long as the overall score
    (as a harmonic mean) of its group relative stake and member
    coverage, is higher than a given threshold.

    Parameters
    ----------
    groups : list of sets
        Resource groups containing resource ids.
    rl : DataFrame
        A resource log.
    p : float
        A given threshold value in range [0, 1.0].

    Returns
    -------
    om : OrganizationalModel
        An organizational model resulted from assigning execution modes 
        to resource groups.
    """
    print('Applying OverallScore with ' +
        'threshold {} '.format(p) + 'for mode assignment:')
    all_execution_modes = set(rl[['case_type', 'activity_type', 'time_type']]
        .drop_duplicates().itertuples(index=False, name=None))

    from orgminer.Evaluation.l2m.diagnostics import \
        group_relative_stake, group_coverage
    from collections import defaultdict

    scores_group_rel_stake = defaultdict(lambda: defaultdict(dict))
    min_score_group_rel_stake = 1.0
    max_score_group_rel_stake = 0.0
    scores_group_cov = defaultdict(lambda: defaultdict(dict))
    min_score_group_cov = 1.0
    max_score_group_cov = 0.0

    # obtain scores
    for i, group in enumerate(groups):
        for m in all_execution_modes:
            rel_stake = group_relative_stake(group, m, rl)
            scores_group_rel_stake[i][m] = rel_stake
            min_score_group_rel_stake = rel_stake \
                if rel_stake < min_score_group_rel_stake \
                else min_score_group_rel_stake
            max_score_group_rel_stake = rel_stake \
                if rel_stake > max_score_group_rel_stake \
                else max_score_group_rel_stake

            cov = group_coverage(group, m, rl)
            scores_group_cov[i][m] = cov
            min_score_group_cov = cov \
                if cov < min_score_group_cov \
                else min_score_group_cov
            max_score_group_cov = cov \
                if cov > max_score_group_cov \
                else max_score_group_cov

    def min_max_scale(value, min_value, max_value):
        return (value - min_value) / (max_value - min_value)
    
    # assign based on threshold filtering
    from operator import itemgetter
    om = OrganizationalModel()
    for i, group in enumerate(groups):
        tmp_modes = list()
        for m in all_execution_modes:
            scaled_rel_stake = min_max_scale(
                scores_group_rel_stake[i][m],
                min_score_group_rel_stake,
                max_score_group_rel_stake)
            scaled_cov = min_max_scale(
                scores_group_cov[i][m],
                min_score_group_cov,
                max_score_group_cov)

            if scaled_rel_stake + scaled_cov > 0:
                relatedness_score = (
                    2 * (scaled_rel_stake * scaled_cov) /
                    (scaled_rel_stake + scaled_cov)
                )
                if relatedness_score > p:
                    tmp_modes.append((m, relatedness_score))
            else:
                pass
        modes = list(item[0] 
            for item in sorted(tmp_modes, key=itemgetter(1), reverse=True))
        om.add_group(group, modes)

    return om

